#include "ell/tp/BarrierTree.hh"
#include "ell/tp/SaddleNet.hh"
#include "ell/minSampling.hh"
#include "ell/WalkAbortionCriterion_.hh"
#include <biu/OptionParser.hh>
#include <ell/rna/S_RNAfe_SingleM.hh>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace biu;
using namespace ell;

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool init(int argc, char** argv, LandscapeTopology*& floodLT,
		LandscapeTopology*& sampleLT, MinimumSelector*& ms,
		SamplingAbortionCriterion*& sac, Walk*& up, Walk*& down,
		string& output, bool& verbose, bool& isSN, double& threshold,
		bool& flood, bool& dynlen, WalkAbortionCriterion*& wac, int& ncsteps,
		int& nctest, size_t& outSteps, size_t& deltaTest, double& delta,
		bool & addNewMin);

//! RNAtpBarrierEstim takes several arguments (see -help) and 2 input files(a SaddleNet produced by flooding
//! and another one produced by sampling) to compute a certain amount of walks on the energy landscape 
//! of a given RNA sequence, and updates a saddle net accordingly with optional output to concole or to file
//! file can than be read by graphviz tools to compute a graph image or given as input
//! for other tools like RNAtpFlood or RNAtpSaddleSample 
int main(int argc, char** argv) {
	string sequence, structure, output;
	bool verbose, isSN, dynlen;

	Walk* up= NULL;
	Walk* down= NULL;
	MinimumSelector* ms= NULL;
	SamplingAbortionCriterion* sac= NULL;
	LandscapeTopology* floodLT= NULL;
	LandscapeTopology* sampleLT= NULL;
	WalkAbortionCriterion* wac= NULL;
	double threshold, delta;
	bool addNewMin = false;
	bool flood;
	int ncsteps, nctest;
	size_t outSteps, deltaTest;
	//! initialize by parsing all given arguments
	//! and init random number generator
	if ( !init(argc, argv, floodLT, sampleLT, ms, sac, up, down, output,
			verbose, isSN, threshold, flood, dynlen, wac, ncsteps, nctest,
			outSteps, deltaTest, delta, addNewMin))
		return -1;
	if (flood) {
		cerr << "flood net has "<<dynamic_cast<BarrierTree*>(floodLT)->getLeafCount()<<" leaves"<<endl;
		cerr << "medium barrier is "<<dynamic_cast<BarrierTree*>(floodLT)->getMediumBarrier()<<endl;
		if (dynamic_cast<BarrierTree*>(floodLT)->check())
			cerr << "alles ok!"<<endl;
		if (verbose) {
			if (isSN)
				cerr << "given Topology is a SaddleNet!"<<endl;
			else
				cerr << "given Topology is a BarrierTree!"<<endl;
			
			cerr << "RMSD between topology "<<floodLT<<" and topology "
					<<sampleLT<<" is: " << floodLT->getRMSD(sampleLT)<<endl;

			cerr << "MinCount floodnet: "<<floodLT->getMinCount()
					<<" and samplenet: " <<sampleLT->getMinCount()<<endl;
		}
	}
	
	std::cout <<"\n minima = " <<sampleLT->getMinCount();
	if (isSN)
		cout << dynamic_cast<SaddleNet*>(sampleLT)->getMediumSaddleVal();
	else
		cout << dynamic_cast<BarrierTree*>(sampleLT)->getMediumBarrier();
	cout <<endl;
	
	//! the REAL work..
	//TODO: wieder aktivieren
	if (dynlen) {
		do {
			static_cast<WAC_MaxLengthSG*>(wac)->setMaxLength((static_cast<WAC_MaxLengthSG*>(wac)->getMaxLength())-1);
			allMinSampling(up, down, sampleLT, ms, sac, threshold, addNewMin);
			if (flood) {
				static_cast<SAC_DeltaNoChange_Output*>(sac)->resetCounter();
			} else {
				static_cast<SAC_DeltaHeight_Output*>(sac)->setStepNum(0);
			}
		} while (static_cast<WAC_MaxLengthSG*>(wac)->getMaxLength() > 1);
		
	} else {
		allMinSampling(up, down, sampleLT, ms, sac, threshold, addNewMin);
	}
	//! write LandscapeTopology to file if -file argument given
	if (output == "stdout") {
		if (isSN)
			cout << *dynamic_cast<SaddleNet*>(sampleLT);
		else
			cout << *dynamic_cast<BarrierTree*>(sampleLT);

	} else if ( !output.empty() ) {
		ofstream fout(output.c_str());

		if (!fout) {
			cerr << "Cannot create/open file \""<<output<<"\"!"<<endl;
			return -1;
		}
		if (isSN)
			fout << *dynamic_cast<SaddleNet*>(sampleLT);
		else
			// TODO geaendert
			fout << *dynamic_cast<BarrierTree*>(sampleLT);

		fout.close();
	}
//	/*TODO: entfernen
//	 cerr << dynamic_cast<BarrierTree*>(floodLT)->getMin(0)->toString()<<endl;
//	 cerr << dynamic_cast<BarrierTree*>(sampleLT)->getMin(0)->toString()<<endl;
//	 */
	//! clean up
	delete up;
	delete down;
	delete ms;
	delete sac;
	if (!flood)
		delete floodLT;
	delete sampleLT;

	return 0;
}

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool init(int argc, char** argv, LandscapeTopology*& _floodLT,
		LandscapeTopology*& _sampleLT, MinimumSelector*& _ms,
		SamplingAbortionCriterion*& _sac, Walk*& _up, Walk*& _down,
		string& _outfile, bool& _verbose, bool& _isSN, double& threshold,
		bool& flood, bool& dynlen, WalkAbortionCriterion*& wac, int& ncsteps,
		int& nctest, size_t& outSteps, size_t& deltaTest, double& delta,
		bool &addNewMin) {

	// init randomizer as an LCG for platform independance	
	RandomNumberGenerator* rng = new RNG_LCG();
	RNF::setRNG(rng);
	delete rng;
	rng=NULL;
	// seed is 1
	const unsigned int seed = 1;
	RNF::getRNG().setSeed(seed);

	// OPTION PARSING
	// the options defined..
	OptionMap allowedArgs;
	allowedArgs.push_back(COption("floodfile", true, COption::STRING,
			"input file which gives the exact saddle net"));
	allowedArgs.push_back(COption(
			"samplefile",
			false,
			COption::STRING,
			"input file which gives an approximated saddle net, if not given an empty saddle net is used"));
	allowedArgs.push_back(COption(
			"output",
			true,
			COption::STRING,
			"if \"stdout\" it writes output to console, else to a file with the given value as name, default is no output"));
	allowedArgs.push_back(COption("steps", false, COption::INT,
			"# of searchs for a minimum(=upwalk + downwalk)"));
	allowedArgs.push_back(COption(
			"delta",
			true,
			COption::DOUBLE,
			"only if steps=0, program aborts when RMSD between exact saddlenet and sampled saddle net is < delta"));
	allowedArgs.push_back(COption(
			"deltatest",
			true,
			COption::INT,
			"only if steps=0, delta!=0, the test for rmsd < delta is only computed every deltatest steps, default is 100",
			"100"));
	allowedArgs.push_back(COption(
			"ncsteps",
			true,
			COption::INT,
			"only if steps=0, program aborts when medium saddle/barrier height did not change for this amount of steps; if floodfile is specified program aborts if RMSD between flood/samplenet did not change for this amount of steps"));
	allowedArgs.push_back(COption(
			"nctest",
			true,
			COption::INT,
			"only if steps=0, ncsteps!=0, the height change test is computed every nctest steps, default is 100",
			"100"));
	allowedArgs.push_back(COption(
			"ms",
			false,
			COption::INT,
			"type of MinimumSelector\n\
1: successive minimum selection\n\
2: uniform random minimum selection\n\
3: bolztmann weighted random minimum selection with factor \'boltz_factor\'"));
	allowedArgs.push_back(COption(
			"up",
			false,
			COption::INT,
			"type of upwards walk: \n\t \
1: RandomWalk with fixed length \n\
2: RandomWalk with random length between length and maxlength"));
	allowedArgs.push_back(COption("down", false, COption::INT,
			"type of downwards walk: \
\n\t 1: Gradient Walk \
\n\t 2: AdaptiveWalk"));
	allowedArgs.push_back(COption("len", false, COption::INT,
			"length of upwards walk or minimal length if up=2"));
	allowedArgs.push_back(COption(
			"dynlen",
			true,
			COption::BOOL,
			"set this for decreasing upwalk length; initial value is set by -len option; criterion for length decrease is set by -ncsteps if no floodnet is specified, or by -delta if floodnet is specified"));
	allowedArgs.push_back(COption(
			"addNewMin",
			true,
			COption::BOOL,
			"if present, new found minima are added to the landscape topology, otherwise not"));
	allowedArgs.push_back(COption("mlen", true, COption::INT,
			"only if up=2, maximum length, default is length+5"));
	allowedArgs.push_back(COption("exp", true, COption::DOUBLE,
			"only if up=2, has impact on random length distribution, default is 1"));
	allowedArgs.push_back(COption(
			"bf",
			true,
			COption::DOUBLE,
			"if ms=3, factor for energy in boltzmann weights, has impact on minimumselection, default is 1"));
	allowedArgs.push_back(COption(
			"out",
			true,
			COption::INT,
			"# of steps after which an output occurs, \
if not specified or 0 it only writes one output at the end, default is 0"));
	allowedArgs.push_back(COption("v", true, COption::BOOL,
			"(v)erbose: if true, it prints out additional info"));
	allowedArgs.push_back(COption("sn", true, COption::BOOL,
			"SaddleNet: if true, uses SaddleNet, else BarrierTree - changes output!"));
	allowedArgs.push_back(COption("Emax", true, COption::DOUBLE,
			"only minima with energy below this value will be counted", "65536"));
	//! info defined..
	COptionParser
			opts =
					COptionParser(
							allowedArgs,
							argc,
							argv,
							"\
			RNATopology: finds minima in the energy landscape for a RNA\n\
			with sequence e.g. AUGCAAU and given nested structure e.g. ((...))");

	if (opts.noErrors()) {
		
		if (opts.argExist("help")) {
			opts.coutUsage();
			return 0;
		}
		
		// temporary variables
		size_t steps/*, outSteps, deltaTest*/;
		size_t length, walkUp_maxLength;
		double walkUp_exp, boltz_factor, /*delta,deltaheight*/;
		ifstream fin, fin2;
		//WalkAbortionCriterion*	wac = NULL;
		//int ncheight, nctest;
		//! set optional Emax argument	    
		threshold = opts.getDoubleVal("Emax");
		//! check for verbose argument
		_verbose = opts.argExist("v");
		//! check for addNewMin argument
		addNewMin = opts.argExist("addNewMin");
		//! check for dynlen argument
		if (opts.argExist("dynlen"))
			dynlen=1;
		else
			dynlen=0;
		//! read floodFile
		if (opts.argExist("floodfile")) {
			fin.open(opts.getStrVal("floodfile").c_str());
			if (!fin)
				cerr << "Cannot open file \""<<opts.getStrVal("floodfile")
						<<"\" for reading SaddleNet!"<<endl;
			if (!fin)
				return false;
			flood=true;
		} else
			flood=false;
		//! read sampleFile
		fin2.open(opts.getStrVal("samplefile").c_str());
		if (!fin2)
			cerr << "Cannot open file \""<<opts.getStrVal("samplefile")
					<<"\" for reading SaddleNet!"<<endl;
		if (!fin2)
			return false;

		if (opts.argExist("sn") && opts.getBoolVal("sn")) {
			_isSN = true;
			_floodLT = new SaddleNet();
			_sampleLT = new SaddleNet();
			if (flood)
				fin >> *dynamic_cast<SaddleNet*>(_floodLT);
			fin2 >> *dynamic_cast<SaddleNet*>(_sampleLT);
		} else {
			_isSN = false;
			_floodLT = new BarrierTree();
			_sampleLT = new BarrierTree();
			if (flood)
				fin >> *dynamic_cast<BarrierTree*>(_floodLT);
			fin2 >> *dynamic_cast<BarrierTree*>(_sampleLT);
		}

		if (opts.argExist("floodfile")) {
			fin.close();
		}
		fin2.close();

		//! check for steps argument
		if (opts.getIntVal("steps") < 0) {
			cerr << "steps has to be >= 0!"<<endl;
			return false;
		} else {
			steps = static_cast<size_t>(opts.getIntVal("steps"));
			delta = 0;

			if (steps != 0 && opts.argExist("delta") && _verbose)
				cout << "delta argument ommited, cause steps != 0"<<endl;
		}

		if (steps==0) {
			if (opts.argExist("delta")) {
				if (opts.getDoubleVal("delta") < 0) {
					cerr << "delta has to be >= 0!"<<endl;
					return false;
				} else if (!flood) {
					cerr
							<<"incompatible abortion criterion! please specify a floodnet (see -floodfile)"
							<<endl;
					return false;
				}

				delta = opts.getDoubleVal("delta");
				if (opts.getIntVal("deltatest") < 0) {
					cerr << "deltatest has to be >= 0!"<<endl;
					return false;
				}
				deltaTest = opts.getIntVal("deltatest");
			}
//			else {
//				cout
//						<< "steps == 0 AND no delta argument means no abortion criterion for sampling, not possible!"
//						<<endl;
//				return false;
//			}
			if (opts.argExist("ncsteps")) {
				if (opts.getIntVal("ncsteps") < 1) {
					cerr<<"ncsteps must be >0!"<<endl;
					return false;
				}
				ncsteps = opts.getIntVal("ncsteps");
				if (opts.getIntVal("nctest") < 0) {
					cerr << "nctest has to be >= 0!"<<endl;
					return false;
				} else
					nctest = opts.getIntVal("nctest");
			}
		}


		//! check for optional outsteps argument	    
		if (opts.argExist("out"))
			if (opts.getIntVal("out") < 0) {
				cerr << "steps has to be >= 0!"<<endl;
				return false;
			} else
				outSteps = static_cast<size_t>(opts.getIntVal("out")) ;
		else {
			//	    	cout << "out set to default 0"<<endl;
			outSteps = 0;
		}
		if (steps > 0)
			_sac = new SAC_MaxCount_Output(steps, cout, outSteps);
		else if (opts.argExist("delta")) {
			if (dynlen == 1) {
				_sac = new SAC_DeltaNoChange_Output(_floodLT, delta, deltaTest, cout, outSteps, ncsteps);
			} else {
				_sac = new SAC_Delta_Output(_floodLT, delta, deltaTest, cout, outSteps);
			}
		} else {
			cerr<<"setting sac to SAC_DeltaHeight_Output..."<<endl;
			_sac = new SAC_DeltaHeight_Output(cout,outSteps,ncsteps,nctest);
		}

		//! check for length argument
		if (opts.getIntVal("len") < 0) {
			cerr << "len has to be at least 0!"<<endl;
			return false;
		} else
			length = static_cast<size_t>(opts.getIntVal("len"));

		switch (opts.getIntVal("ms")) {
		case 1:
			_ms = new MS_Successive();
			break;
		case 2:
			_ms = new MS_Uniform();
			break;
		case 3:
			if (opts.argExist("bf")) {
				boltz_factor = opts.getDoubleVal("bf");
				if (boltz_factor<1) {
					cerr << "argument bf has to be >=1 !"<<endl;
					return false;
				}
			} else
				boltz_factor = 1;

			_ms = new MS_Boltzmann(boltz_factor);
			break;
		default:
			//TODO: don't forget to refresh the error message when # of options changed
			cerr << "ms has to be element of [1,2,3]";
			return false;
		}// switch(ms)


		//! check for up argument
		switch (opts.getIntVal("up")) {
		case 1:
			if (dynlen == 1) {
				wac = new WAC_MaxLengthSG(length+1);
			} else
				wac = new WAC_MaxLength(length);
			break;
		case 2:
			if (opts.argExist("mlen")) {
				int walkUp_maxLength_ = opts.getIntVal("mlen");
				if (walkUp_maxLength_ <= (int)length) {
					cerr << "mlen has to be greater than len!"<<endl;
					return false;
				} else
					walkUp_maxLength = static_cast<size_t>(walkUp_maxLength_);
			} else
				walkUp_maxLength = length + 5;

			if (opts.argExist("exp")) {
				walkUp_exp = opts.getDoubleVal("exp");
				if (walkUp_exp < 0) {
					cerr << "exp has to be at least 0!"<<endl;
					return false;
				}
			} else
				walkUp_exp = 1;

			if (walkUp_exp == 1)
				wac = new WAC_Random_Linear(length, walkUp_maxLength);
			else
				wac = new WAC_Random_Exp(length, walkUp_maxLength, walkUp_exp);

			break;
		default:
			//TODO: don't forget to refresh the error message when # of options changed
			cerr << "up has to be elemt of [1,2]!";
			return false;
		}// switch(up)
		_up = new WalkRandom(wac);

		//! check for down argument
		switch (opts.getIntVal("down")) {
		case 1:
			_down = new WalkGradient(true);
			break;
		case 2:
			_down = new WalkAdaptive(WalkAdaptive::RANDOM, true);
			break;
		default:
			//TODO: don't forget to refresh the error message when # of options changed
			cerr << "down has to be element of [1,2]!";
			return false;
		}// switch(down)

		//! check for outfile argument
		if (opts.argExist("output"))
			_outfile = opts.getStrVal("output");

	} // if (opts.noErrors())
	else
		return false;

	//parsing successfull
	return true;
}
