
#include <iostream>
#include <fstream>
#include <string>

#include <biu/OptionParser.hh>

#include <ell/rna/S_RNAfe_SingleM.hh>

#include "ell/tp/BarrierTree.hh"
#include "ell/tp/SaddleNet.hh"
#include "ell/minSampling.hh"
#include "ell/WalkAbortionCriterion_.hh"
#include "ell/directWalk.hh"

using namespace std;
using namespace biu;
using namespace ell;

enum TP_TYPES {BT, SN};


void printStatistics(	std::ostream& out,
						const size_t & numOfWalks, 
						const LandscapeTopology * const lt,
						const TP_TYPES & ltType)
{
	out	<<numOfWalks <<"\t" <<lt->getMinCount() <<"\t";
	switch (ltType) {
	case BT : 
		out <<(static_cast<const BarrierTree* const>(lt))->getMediumBarrier();
		break;
	case SN :
		out <<(static_cast<const SaddleNet* const>(lt))->getMediumSaddleVal();
		break;
	default :
		std::cerr <<"\nINTERNAL ERROR : given TP_TYPE not handled for statistics !\n";
		out <<"?";
	}
	out <<"\n";
}

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
//bool init(int argc, char** argv, LandscapeTopology*& floodLT,
//		LandscapeTopology*& sampleLT, MinimumSelector*& ms,
//		SamplingAbortionCriterion*& sac, Walk*& up, Walk*& down,
//		string& output, bool& verbose, bool& isSN, double& threshold,
//		bool& flood, bool& dynlen, WalkAbortionCriterion*& wac, int& ncsteps,
//		int& nctest, size_t& outSteps, size_t& deltaTest, double& delta);

void initArguments(OptionMap & allowedArgs, std::string &infoString) {
	allowedArgs.push_back(COption(
			"inTP",
			true,
			COption::STRING,
			"file name of the initial landscape topology, if 'STDIN' it is read from standard input",
			"STDIN"));
	allowedArgs.push_back(COption(
			"inType",
			true,
			COption::STRING,
			"type of the initial landscape topology read from 'inTP' (SN,BT)",
			"SN"));
	allowedArgs.push_back(COption(
			"outTP",
			true,
			COption::STRING,
			"file name of the result landscape topology, if 'STDOUT' it is written to standard output",
			"STDOUT"));
	allowedArgs.push_back(COption(
			"outType",
			true,
			COption::STRING,
			"type of the result landscape topology to write to 'outTP' (BT)",
			"BT"));
	allowedArgs.push_back(COption(
			"walks",
			true,
			COption::INT,
			"number of walks between randomly choosen minima to do"));
	allowedArgs.push_back(COption(
			"ncsteps",
			true,
			COption::INT,
			"stop if no barrier decrease after the given number of walks between randomly choosen minima can be seen"));
	allowedArgs.push_back(COption(
			"mhIter",
			true,
			COption::INT,
			"iterations of Morgan-Higgs random direct path sampling for each barrier",
			"10"));
	allowedArgs.push_back(COption(
			"mhAvgBmin",
			true,
			COption::DOUBLE,
			"stop if the average barrier energy is below the given value"));
	allowedArgs.push_back(COption(
			"outEach",
			true,
			COption::INT,
			"output some statistics after each given number of direct walks"));
	allowedArgs.push_back(COption(
			"v",
			true,
			COption::BOOL,
			"verbose output"));
	allowedArgs.push_back(COption(
			"help",
			true,
			COption::BOOL,
			"help output of usage and parameters"));
	
	infoString = std::string("Estimates all barriers between all pairs of minima")
				+std::string(" using a direct path sampling introduced by Morgan")
				+std::string(" and Higgs.")
				+std::string("\n")
				+std::string("Each time two minima are randomly choosen, the direct")
				+std::string(" walk is calculated and the energy barrier between ")
				+std::string(" these barriers is updated if lower than already known.")
				;
}

//! RNAtpBarrierEstim takes several arguments (see -help) and 2 input files(a SaddleNet produced by flooding
//! and another one produced by sampling) to compute a certain amount of walks on the energy landscape 
//! of a given RNA sequence, and updates a saddle net accordingly with optional output to concole or to file
//! file can than be read by graphviz tools to compute a graph image or given as input
//! for other tools like RNAtpFlood or RNAtpSaddleSample 
int main(int argc, char** argv) {

	////////////////////////////  INIT PARAMETERS  /////////////////////////
	
	OptionMap allowedArgs;
	std::string infoString;
	initArguments(allowedArgs, infoString);	// init the allowed program arguments
	
	COptionParser opts (	allowedArgs,	// the option parser
							argc,
							argv,
							infoString) ;
	
	  // stop if not parsable
	if (!opts.noErrors()) {
		return -1;
	} else if (opts.argExist("help")) {	// help output
		opts.coutUsage();
		return 0;
	}
	
	////////////////////////////  DEFINE VARIABLES  /////////////////////////

	bool verbose = false;
	TP_TYPES in_type = SN;				// type of the LT to read
	TP_TYPES out_type = BT;				// type of the LT to write
	LandscapeTopology* lt = NULL;		// the LT to change
	size_t mh_iterations = 1;			// iterations of Morgan-Higgs random direct path sampling for each barrier
	std::ostream* output = &std::cout;	// output stream to write the final LT to
	double avgBarrierEmin = -((double)UINT_MAX);	// stop estimation if the average barrier height is lower than this
	bool checkAvgBarrierE = false;		// whether the average barrier height has to be checked
	size_t outEach = 0;					// give some statistics each number of steps
	size_t maxWalks = 0;				// maximal number of walks to do 
	size_t ncSteps = 0;					// maximal number of walks without barrier change to do 
	
	////////////////////////////  PARSE PARAMETERS  /////////////////////////

	verbose = opts.argExist("v");
	
	//######## landscape topolgy type to read
	if (opts.getStrVal("inType").compare("SN") == 0) {
		in_type = SN;
	} else if (opts.getStrVal("inType").compare("BT") == 0) {
		in_type = BT;
	} else {
		std::cerr	<<"\nERROR : given input landscape topology type '" 
					<<opts.getStrVal("inType")
					<<"' is not supported !\n";
		return -1;
	}
	
	//######## landscape topolgy type to write
	if (opts.getStrVal("outType").compare("BT") == 0) {
			out_type = BT;
//	} else if (opts.getStrVal("outType").compare("SN") == 0) {
//		out_type = SN;
	} else {
		std::cerr	<<"\nERROR : given output landscape topology type '" 
					<<opts.getStrVal("outType")
					<<"' is not supported !\n";
		return -1;
	}

	//######## minimal average barrier height to reach for an early stop
	if (opts.argExist("mhAvgBmin")) {
		if (out_type != BT) {
			std::cerr <<"\nERROR : minimal average energy of barrier height only for barrier tree output available !\n";
			return -1;
		}
		checkAvgBarrierE = true;
		avgBarrierEmin = opts.getDoubleVal("mhAvgBmin");
	}

	//######## MH iterations
	if (opts.getIntVal("mhIter") < 1) {
		std::cerr <<"\nERROR : given iterations '" <<opts.getIntVal("mhIter")
					<<"' is less than 1 !\n";
		return -1;
	} else {
		mh_iterations = (size_t)opts.getIntVal("mhIter"); 
	}
	
	//######## stop criteria check
	if (!opts.argExist("walks") && ! opts.argExist("ncsteps")) {
		std::cerr <<"\nERROR : neither 'walks' nor 'ncsteps' given --> no abortion criterion !\n";
		return -1;
	}
	if (opts.argExist("walks") && opts.argExist("ncsteps")) {
		std::cerr <<"\nERROR : either 'walks' OR 'ncsteps' can be specified !\n";
		return -1;
	}
	
	//######## stop criteria : max steps
	if (opts.argExist("walks")) {
		if (opts.getIntVal("walks") < 1) {
			std::cerr <<"\nERROR : given maximal number of walks '" <<opts.getIntVal("walks")
						<<"' is less than 1 !\n";
			return -1;
		} else {
			maxWalks = (size_t)opts.getIntVal("walks"); 
		}
	}
	
	//######## stop criteria : no change steps
	if (opts.argExist("ncsteps")) {
		if (opts.getIntVal("ncsteps") < 1) {
			std::cerr <<"\nERROR : given maximal number of walks without change '" <<opts.getIntVal("ncsteps")
						<<"' is less than 1 !\n";
			return -1;
		} else {
			ncSteps = (size_t)opts.getIntVal("ncsteps"); 
		}
	}
	
	//######## statistics output
	if (opts.argExist("outEach")) {
		if (opts.getIntVal("outEach") < 1) {
			std::cerr <<"\nERROR : given statistics output steps '" <<opts.getIntVal("outEach")
						<<"' is less than 1 !\n";
			return -1;
		} else {
			outEach = (size_t)opts.getIntVal("outEach"); 
		}
	}
	
	//######## read input topology
	if (opts.getStrVal("inTP").size() != 0)
	{
		if (verbose)	
			std::cout <<"\n =======  READ INPUT LANDSCAPE  ========\n" <<std::endl;
		std::istream * input = &std::cin;
		std::ifstream * infile = NULL;
		LandscapeTopology* in_lt = NULL;		// the LT to fill
		if (opts.getStrVal("inTP").compare("STDIN") != 0) {
			infile = new std::ifstream(opts.getStrVal("inTP").c_str());
			input = dynamic_cast<std::istream*>(infile);
		}
		if (input != NULL && input->good()) {
			switch (in_type) {
			case BT : 
					  // read barrier tree from input
					in_lt = new BarrierTree();
					*input >> *(dynamic_cast<BarrierTree*>(in_lt));
					break;
			case SN : 
					  // read saddle network from input
					in_lt = new SaddleNet();
					*input >> *dynamic_cast<SaddleNet*>(in_lt);
					break;
			default :
					std::cerr <<"\nERROR : unknown input LT_TYPE '" <<in_type <<"' !\n";
					return -2;
			}
		} else {
			std::cerr <<"\nERROR : cannot read landscape input '"<<opts.getStrVal("inTP") <<"' !\n";
			return -1;
		}
		if (infile != NULL) {  // clear memory
			delete infile;
		}
		assert(in_lt != NULL);
		
		//######## create output landscape topology type to write
		
		// do conversion
		if (in_type != out_type) {
			switch (out_type) {
			case BT : 
					lt = new BarrierTree();
					break;
			case SN : 
					lt = new SaddleNet();
					break;
			default :
					std::cerr <<"\nERROR : unknown output LT_TYPE '" <<out_type <<"' !\n";
					return -2;
			}
			  // copy all minima
			for (size_t i=0; i<in_lt->getMinCount(); i++)
				lt->addMin(in_lt->getMin(i));
			if (verbose) {
				std::cout	<<"\n Different LT input and output type : only minima copied !" 
							<<std::endl;
			}
			  // remove input landscape because it was copied;
			delete in_lt;
		} else {
			delete lt;
			lt = in_lt;
		}
		if (verbose)	
			std::cout <<"\n =======================================\n" <<std::endl;
	} else {
		std::cerr <<"\nERROR : no input landscape given !\n";
		return -1;
	}
	
	//######## open output stream
	if (opts.getStrVal("outTP").compare("STDOUT") != 0) {	// file name given
		if (opts.getStrVal("outTP").size() == 0) {
			std::cerr <<"\nERROR : no output file name given !\n";
			return -1;
		}
		output = new std::ofstream(opts.getStrVal("outTP").c_str());
	}
	if (output->bad()) {
		std::cerr <<"\nERROR : cannot open outstream '" <<opts.getStrVal("outTP") <<"' !\n";
		return -1;
	}

	
	/////////////////////  PARAMETER SETTING  ////////////////////////////
	if (verbose) {
		std::cout
					<<"\n LT input         : " <<opts.getStrVal("inTP")
					<<"\n LT input type    : " <<opts.getStrVal("inType")
					<<"\n LT output        : " <<opts.getStrVal("outTP")
					<<"\n LT output type   : " <<opts.getStrVal("outType")
					<<"\n MH iterations    : " <<opts.getStrVal("mhIter")
					;
		if (opts.argExist("mhAvgBmin")) {
			std::cout 
					<<"\n MH barrier min E : " <<opts.getStrVal("mhIter");
		}
		std::cout
					<<std::endl;
					
	}
	
	///////////////////  DO BARRIER ESTIMATION  ////////////////////////////
	if (verbose)	
		std::cout <<"\n =========  GET MH ESTIMATION ==========\n" <<std::endl;

	
	size_t numOfWalks = 0; //< number of MH walks done so far
	size_t numOfNCWalks = 0; //< number of walks without a barrier change
	bool anotherWalk = true;  //< stop criterion for while loop
	size_t minFrom = 0, minTo = 0; //< minima indices that will be randomly choosen
	unsigned int numberOfMinima = lt->getMinCount();
	do {
		  // select randomly 2 minima
		minFrom = (size_t)biu::RNF::getRN(numberOfMinima);
		minTo = (size_t)biu::RNF::getRN(numberOfMinima-1);
		minTo += (minTo < minFrom) ? 0 : 1 ; // ensure that minTo != minFrom
		
		  // get barrier estimation via Morgan-Higgs
		SC_Maximum sc_max;
		directWalk( lt->getMin(minFrom), lt->getMin(minTo), &sc_max, mh_iterations);
		numOfWalks++;
		
		  // write estimation to LT
		bool barrierChanged = lt->addSaddle( minFrom, minTo, sc_max.getMaximum() );
		
		  // print statistics
		if ( outEach > 0 && numOfWalks % outEach == 0 ) {
			printStatistics( std::cout, numOfWalks, lt, out_type);
		}
		
		  // update data if a barrier was changed
		if (barrierChanged) {
			numOfNCWalks = 0;
			numberOfMinima = lt->getMinCount();
		} else {
			numOfNCWalks++;
		}
		
		  // check for abortion
		if (maxWalks == 0) {
			anotherWalk = numOfNCWalks < ncSteps;
		} else {
			anotherWalk = numOfWalks < maxWalks;
		}
		if (	anotherWalk
				&& barrierChanged
				&& checkAvgBarrierE ) 
		{
			anotherWalk = (dynamic_cast<BarrierTree*>(lt))->getMediumBarrier() >= avgBarrierEmin;
		}
	} while (anotherWalk);
		
	if (outEach > 0) {
		printStatistics( std::cout, numOfWalks, lt, out_type);
		std::cout.flush();
	}
	

	////////////////////////  WRITE OUTPUT  ////////////////////////////////
	if (verbose)	
		std::cout <<"\n =========  WRITE OUTPUT ==========\n" <<std::endl;
	
	switch (out_type) {
	case BT : // barrier tree
			*output << *(dynamic_cast<BarrierTree*>(lt));
			break;
	case SN : // saddle network 
			*output << *dynamic_cast<SaddleNet*>(lt);
			break;
	default :
			std::cerr <<"\nERROR : unknown output LT_TYPE '" <<out_type <<"' !\n";
			return -2;
	}
	output->flush();
	
	////////////////////////  CLEAR MEMORY  ////////////////////////////////
	
	if (opts.getStrVal("outTP").compare("STDOUT") != 0) {
		(static_cast<std::ofstream*>(output))->close();
		delete output;
	}
	if (lt != NULL) delete lt;

} // end of main

