#include <iostream>
#include <fstream>
#include <string>
#include <biu/OptionParser.hh>
#include <biu/RandomNumberFactory.hh>
#include <ell/rna/S_RNAfe_SingleM.hh>
#include <ell/SC_Counting.hh>
#include "ell/tp/SaddleNet.hh"
#include "ell/TopologyFlooder.hh"
#include "ell/SPC_Counting.hh"

using namespace std;
using namespace biu;
using namespace ell;

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool 	init(	int argc,
				char** argv,
				string& infile, 
				string& outfile,
				double& maxE,
				double& deltaE,
				size_t& maxToStore,
				bool& verbose); 

//! RNAtpFlood takes several arguments (see -help) to flood an energylandscape of RNA 
//! this landscape is defined a RNA sequence and by some known local minima(given by input file)
//! it floods to Emax and updates the given SaddleNet with exact saddleheights up to Emax(but not with new minima!)
//! after completion it writes the updated saddlenet to an output file
int main( int argc, char** argv ) {
	biu::Matrix<StatePairCollector*> mSPC_C;
	std::vector<StateCollector*> vSC;
	
	string 	inputFile, outputFile;
	double	maxE, deltaE;
	size_t	maxToStore;
	
	bool	verbose;
	
	ifstream in;
	ofstream out;
		
	//! init all variables and check whether parameters are all ok (e.g. file names correct)
	if( !init(argc, argv, inputFile, outputFile, maxE, deltaE, maxToStore, verbose))
		return -1;

	in.open(inputFile.c_str());
	if( !in ) { 
		cout << "Cannot open file \""<<inputFile<<"\" for reading SaddleNet!"<<endl; 
		return false; 
	}

	//! creating SaddleNet
	SaddleNet* sn = new SaddleNet();
	//! reading SaddleNet
	if(verbose)	cout << "read SaddleNet from file \""<<inputFile<<"\" ..."; 
	in >> *sn;
	in.close();
	if(verbose)	cout <<"DONE! has "	<< sn->getMinCount()
				<<" minima and "<<sn->getSaddleCount()<<" saddles."<<endl;
	
	//! filling vector of State Collectors for all known minima
	for (size_t i = 0; i < sn->getMinCount(); ++i) {
		//! we decide to only collect and count the states of each basin
		vSC.push_back(new SC_Counting());
	}

	if(verbose)	cerr << "flooding up to maximum energy "<<maxE
				<<", delta="<<deltaE<<", buffer="<<maxToStore<<" ... ";
	//! flooding..
	TopologyFlooder<SPC_Counting>::floodAll(sn, vSC, mSPC_C, maxE, deltaE, maxToStore);
	if(verbose)	cout <<"DONE!"<<endl
				<<"new saddlenet has "<< sn->getMinCount()
				<<" minima and "<<sn->getSaddleCount()<<" saddles."<<endl;
		
	out.open(outputFile.c_str());
	if( !out ) { 
		cout << "Cannot create/open file \""<<outputFile<<"\" for writing updated SaddleNet!"<<endl; 
		return false; 
	} 
	
	if(verbose)	cout << "writing new saddle net to file \""<<outputFile<<"\" ... ";
	out << *sn;
	out.close();
	if(verbose)	cout <<"DONE!"<<endl;
	
	if(verbose)	cout << "deleting all pointers ... ";
	//! clear everything
	for (size_t i = 0; i < sn->getMinCount(); ++i)
		for (size_t j = i; j < sn->getMinCount(); ++j)
			delete mSPC_C[i][j];

	for (size_t i = 0; i < sn->getMinCount(); ++i)
		delete vSC[i];
	
	delete sn;
	if(verbose)	cout <<"DONE! program ended."<<endl;		
	return 0;
}


//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool init(	int argc, 
			char** argv, 
			string& _infile, 
			string& _outfile,
			double& _maxE,
			double& _deltaE,
			size_t& _maxToStore,
			bool& 	_verbose)			
{
	
	// init randomizer as an LCG for platform independance	
	RandomNumberGenerator* rng = new RNG_LCG();
	RNF::setRNG( rng );
	delete rng;
	rng=NULL;
	// seed is 1
	const unsigned int seed = 1;	
	RNF::getRNG().setSeed(seed);

	// OPTION PARSING
	// the options defined..
	OptionMap allowedArgs;
	allowedArgs.push_back(COption("infile", false, COption::STRING, "name of file in which SaddleNet is stored(e.g. output of RNAtpMinSample)"));
	allowedArgs.push_back(COption("E", false, COption::DOUBLE, "maximum energy level the landscape is flooded"));
	allowedArgs.push_back(COption("delta", false, COption::DOUBLE, "value by which e is decreased if state buffer is full"));
	allowedArgs.push_back(COption("buffer", false, COption::INT, "number of states the queue can handle, should be as high as neccessary"));
	allowedArgs.push_back(COption("outfile", true, COption::STRING, "name of file in which to store the flooded SaddleNet, default is floodnet.txt"));
	allowedArgs.push_back(COption("v", true, COption::BOOL, "(v)erbose, gives additional info if set, default is false"));
	
	// info defined..
	COptionParser opts = COptionParser(allowedArgs, argc, argv, "\
			RNAtpFlood: uses given minima in the energy landscape for \n\
			flooding the landscape and updating the given Saddlenet(but not updating new found minima!)");
	
	// temporary variables
	string sequence, structure, infile;
	ifstream in;
	ofstream out;	

	// getting the parameter references filled	
	if (opts.noErrors()) {
  		_infile = opts.getStrVal("infile");
 		   
	    _maxE = opts.getDoubleVal("E");
	    
	    if(opts.getDoubleVal("delta") <= 0) {
	    	cerr << "delta has to be >0 !"<<endl;
	    	return false;
	    } else
	    	_deltaE = opts.getDoubleVal("delta");
	    
	    if(opts.getIntVal("buffer") < 10000) {
	    	cerr << "buffer has to be at least 10000 !"<<endl;
	    	return false;
	    } else 
	    	_maxToStore = opts.getIntVal("buffer");
	    
	    //! check for verbose argument
	    if(opts.argExist("v"))
	    	_verbose = opts.getBoolVal("v");
	    else
	    	_verbose = false;
	    
	    //! check for outfile argument
	    if(opts.argExist("outfile"))
	    	_outfile = opts.getStrVal("outfile");
	    else
	    	_outfile = "floodnet.txt";
  
	}  // if (opts.noErrors())
	else return false;
	
	//parsing successfull
	return true;
}

