#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <biu/OptionParser.hh>
#include <ell/rna/S_RNAfe_SingleM.hh>
#include "ell/tp/MinimaSet.hh"

using namespace std;
using namespace biu;
using namespace ell;

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool 	init(	int argc,
				char** argv,
				string& infile, 
				string& outfile,
				bool& verbose); 

//! RNAtpFlood takes several arguments (see -help) to flood an energylandscape of RNA 
//! this landscape is defined a RNA sequence and by some known local minima(given by input file)
//! it floods to Emax and updates the given SaddleNet with exact saddleheights up to Emax(but not with new minima!)
//! after completion it writes the updated saddlenet to an output file
int main( int argc, char** argv ) {

	clock_t start, end;
	string 	inputFile, outputFile;
	bool	verbose;
	
	ifstream fin, fin2;
	ofstream fout;
		
	//! init all variables and check whether parameters are all ok (e.g. file names correct)
	if( !init(argc, argv, inputFile, outputFile, verbose))	return -1;

	fin.open(inputFile.c_str());
	if( !fin ) { 
		cerr << "Cannot open file \""<<inputFile<<"\" for reading all conformations!"<<endl; 
		return false; 
	}

	//! creating MinimaSet
	MinimaSet* mn = new MinimaSet();
	//! reading MinimaSet
	if(verbose)	cerr << "read MinimaNet from file \""<<inputFile<<"\" ...";
	start = clock();
	mn->readSuboptFile(fin);
	end = clock();
	if(verbose)	cerr <<"DONE! has "	<< mn->getMinCount()<<" minima, took"
					<<static_cast<double>(end-start)/CLOCKS_PER_SEC<<" seconds CPU time."<<endl;
	fin.close();
	
	if(verbose)	cerr << "now checking every minima in MinimaNet if it has a neighbor with lower/even energy ...";

	bool isMin, isPotShoulder;
	size_t counter= 0, counter2=0;
	const State* s = mn->getMin(0);
	State::NeighborListPtr neighbors = s->getNeighborList(); //pointer to a list off all neighbors 	
	State::NeighborList::Iterator nl_it = neighbors->begin();
	
	start = clock();
	for (size_t i = 0; i < mn->getMinCount(); ++i) {
		s = mn->getMin(i);
		neighbors = s->getNeighborList();
		nl_it = neighbors->begin();
		isMin = true;
		isPotShoulder = false;
//		if(i==1) cerr << "Min 1 is "<< s->toString()<<" and neighbors are.."<<endl;
		
		for (nl_it = neighbors->begin(); nl_it != neighbors->end(); ++nl_it) {
//			if(i==1)
//				cerr<< nl_it->toString()<<endl;
			if(nl_it->getEnergy() == s->getEnergy())
				isPotShoulder=true;
		
			if(nl_it->getEnergy() < s->getEnergy())
				isMin=false;
		}
		if(isMin)  ++counter;
		if(isMin && isPotShoulder) ++counter2;

	}
	end = clock();
	if(verbose) cerr <<"DONE! took "<<static_cast<double>(end-start)/CLOCKS_PER_SEC<<" seconds CPU time."<<endl;

	cerr << "found "<<counter<<" true minima out of "
	     <<mn->getMinCount()<<" given states, "<<counter2<<" of them potential shoulder points or valley"<<endl;
	
	delete mn;
	if(verbose)	cerr <<"DONE! program ended."<<endl;		
	return 0;
}


//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool init(	int argc, 
			char** argv, 
			string& _infile, 
			string& _outfile,
			bool& 	_verbose)			
{
	
	// OPTION PARSING
	// the options defined..
	OptionMap allowedArgs;
	allowedArgs.push_back(COption("infile", false, COption::STRING, "input file with format of RNAsubopt"));
	allowedArgs.push_back(COption("outfile", true, COption::STRING, "output file with native format"));
	allowedArgs.push_back(COption("v", true, COption::BOOL, "(v)erbose, gives additional info if set, default is false"));
	
	// info defined..
	COptionParser opts = COptionParser(allowedArgs, argc, argv, "\
			RNAtpFlood: uses given minima in the energy landscape for \n\
			flooding the landscape and updating the given Saddlenet(but not updating new found minima!)");
	
	// temporary variables
	string infile;
	ifstream in;


	// getting the parameter references filled	
	if (opts.noErrors()) {
  		_infile = opts.getStrVal("infile");
  		
  		if(opts.argExist("outfile"))
  			_outfile = opts.getStrVal("outfile");
	    //! check for verbose argument
	    if(opts.argExist("v"))
	    	_verbose = opts.getBoolVal("v");
	    else
	    	_verbose = false;
  
	}  // if (opts.noErrors())
	else return false;
	
	//parsing successfull
	return true;
}

