#include "ell/tp/MinimaSet.hh"
#include "ell/minSampling.hh"
#include "ell/WalkAbortionCriterion_.hh"
#include "ell/SamplingAbortionCriterion.hh"
#include <biu/OptionParser.hh>
#include <ell/rna/S_RNAfe_SingleM.hh>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace biu;
using namespace ell;


#include <ViennaRNA/fold_vars.h>
extern "C"	float fold(const char *sequence, char *structure);

enum TP_TYPE {MS,BT,SN};

//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool 	
init(	int argc, char** argv,
		string& sequence, string& structure,
		MinimumSelector*& ms, SamplingAbortionCriterion*& sac, WalkAbortionCriterion*& _wacup,
		Walk*& up, Walk*& down,
		string& outfile, bool& verbose, 
		int& iterations, int& outSteps, size_t& sac_type, size_t& min, 
		double& Emax, TP_TYPE & outType,
		LandscapeTopology *& LT);

//! RNAtpMinSample takes several arguments (see -help) to compute a certain amount of walks
//! on the energylandscape of a given RNA sequence, given by a starting RNA structuresequence(which has to be a local minimum)
//! and updates a saddle net accordingly with optional output to concole or to file
//! file can than be read by graphviz tools to compute a graph image or given as input
//! for other tools like RNAtpFlood or RNAtpSaddleSample 
int main( int argc, char** argv ) {
	string	sequence, structure, outFile;
	bool verbose;
	Walk* up = NULL;
	Walk* down= NULL;
	MinimumSelector* ms	= NULL;
	SamplingAbortionCriterion* 	sac = NULL;
	WalkAbortionCriterion* _wacup = NULL;
//	int stepinc, iter;
	int iterations, outSteps;
	size_t min, sac_type;
	double Emax;
	TP_TYPE outType = MS;

	//! LandscapeTopology to fill
	LandscapeTopology*	LT	= NULL;
	
	//! initialize by parsing all given arguments
	//! and init random number generator
	if( !init(argc, argv, sequence, structure, ms, sac, _wacup, up, down, outFile, verbose,  iterations, outSteps, sac_type, min, Emax, outType, LT)) 
		return -1;
	
	//! we work on RNA with moveset of single moves..
	S_RNAfe_SingleM*	rna = new S_RNAfe_SingleM(sequence,structure);

	//! check if given RNA is local minimum via a gradient walk
	StateCollector*	 scc = new SC_Counting();
	WalkGradient::walkGradient(rna,scc,true);
	//! if gradient walk from given RNA is longer than 1(1 because of start state)
	//! the given RNA is no local minimum -> abortion
	if(scc->size() > 1) {
		cerr << "\nERROR: given RNA is no local minimum! aborted.."<<endl;
		delete LT;	delete rna;	delete ms;	delete up;	delete down; delete scc;
		return -1;
	}
	
	//! add given RNA from which we know it's a local minimum
	if (LT->getMinIndex(rna) == INT_MAX) { // not already present in LT
		LT->addMin(rna);
	}

	// sample minima 
	allMinSampling(up,down,LT,ms,sac, Emax);
	
	  // do line break if sampling statistics have been printed
	if (outSteps != 0) {
		std::cout <<std::endl;
	}
		
	// do output
	if(verbose)	
		cout 	<<"\n"
				<< " -> Writing topology to '"<<outFile<<"' ... " <<std::endl;
	std::ostream* output = &std::cout;
	if(outFile.compare("STDOUT") != 0) {
		ofstream * fout = new ofstream(outFile.c_str());
		if( fout->bad() ) { 
			cerr << "Cannot create/open file \""<<outFile<<"\" for writing resulting landscape topology!\n";
			delete fout;
			return -1; 
		} 
		output = fout;
	}
	  // force landscape type specific output
	switch(outType) {
	case MS :
		*output << *static_cast<MinimaSet*>(LT);
		break;
	case BT :
		*output << *static_cast<BarrierTree*>(LT);
		break;
	case SN :
		*output << *static_cast<SaddleNet*>(LT);
		break;
	default :
		std::cerr <<"\nERROR : given landscape topology type is not handled !\n";
		return -1;
	}
	  // empty stream buffer
	output->flush();
	  // close file if necessary
	if(outFile.compare("STDOUT") != 0) {
		static_cast<std::ofstream*>(output)->close();
		delete output;
		output = &std::cout;
	}
	
	
	//! clean up
	delete scc;
	delete rna;
	delete LT;
	delete up;
	delete down;
	delete ms;
	delete sac;

	return 0;
}





//! init function for parsing the arguments given to the binary
//! @return true if parsing of arguments was successfull
bool init(	int argc, 
			char** argv, 
			string& _sequence, 
			string& _structure, 
			MinimumSelector*& _ms, 
			SamplingAbortionCriterion*& _sac, 
			WalkAbortionCriterion*& _wacup,
			Walk*& _up, 
			Walk*& _down, 
			string& _outfile,
			bool& _verbose,
			int& iterations, 
			int& outSteps,
			size_t& sac_type,
			size_t& minMax,
			double& Emax, 
			TP_TYPE & outType,
			LandscapeTopology *& LT
			) 
{
	
	// OPTION PARSING
	// the options defined..
	OptionMap allowedArgs;
	allowedArgs.push_back(COption("seq", false, COption::STRING, "RNA sequence e.g.  'AUCCUG' of that the landscape topology should be explored"));
	allowedArgs.push_back(COption("str", true, COption::STRING, "RNA structure e.g. '(...).', if none is given, the minimum free energy structure is used"));
	allowedArgs.push_back(COption("iter", false, COption::INT, "number iterations to find a new minimum"));
	allowedArgs.push_back(COption("len", false, COption::INT, "maximal length of upwards walk"));
	allowedArgs.push_back(COption("ncsteps", true, COption::INT, "only if iter=0, program aborts when this # of iterations occured without change to # of minima"));
	allowedArgs.push_back(COption("outTP", true, COption::STRING, "file name to write the resulting landscape topology to. If equal to 'STDOUT' it is written to the standard output.","STDOUT"));
	allowedArgs.push_back(COption("outType", true, COption::STRING, "type of the landscape topology to sample and write to 'outTP' {MS,BT,SN}","MS"));
	allowedArgs.push_back(COption("inTP", true, COption::STRING, "extend the landscape topology from this file. If equal to 'STDIN' it is read from the standard input."));
	allowedArgs.push_back(COption("ms", true, COption::INT, "type of MinimumSelector \n1: successive minimum selection (final min of last walk) \n2: uniform random minimum selection \n3: boltzmann weighted random minimum selection using factor \'boltz_factor\'","2"));
	allowedArgs.push_back(COption("up", true, COption::INT, "type of upwards part of the walk: \n\t 1: RandomWalk with fixed length \n2: RandomWalk with random length between len and lenMin","1"));
	allowedArgs.push_back(COption("down", true, COption::INT, "type of downwards walk: \n\t 1: Gradient Walk \n\t 2: AdaptiveWalk","2"));
	allowedArgs.push_back(COption("minMax", true, COption::DOUBLE, "abort after having found this  number of minima" ));
	allowedArgs.push_back(COption("lenMin", true, COption::INT, "only if up=2 (variable length), minimal length of the upwards walk","1"));
	//TODO needs further testing
	allowedArgs.push_back(COption("Emax", true, COption::DOUBLE, "only minima with energy below this value will be counted", "65536"));
	allowedArgs.push_back(COption("randSeed", true, COption::INT, "initialize random seed", "1"));
//	allowedArgs.push_back(COption("exp", true, COption::DOUBLE, "only if up=2, has impact on random length distribution","1"));
	allowedArgs.push_back(COption("bf", true, COption::DOUBLE, "if ms=3, factor for energy in boltzmann weights, has impact on minimumselection","1"));
	allowedArgs.push_back(COption("outEach", true, COption::INT, "# of iterations after which an output of the sampling statistic occurs, \
if not specified or 0 no output is generated","0"));
	allowedArgs.push_back(COption("v", true, COption::BOOL, "do verbose output."));
	
	
	// info defined..
	COptionParser opts = COptionParser(allowedArgs, argc, argv, "\
 Finds minima in the energy landscape of an RNA molecule\
 with given sequence and initial nested structure.\
 ");
	
	// temporary variables
	int 		 	seed;
	size_t 		length, walkUp_minLength;
	double		boltz_factor;
	size_t		noChangeSteps = 0;
//	double		walkUp_exp;
	

	// getting the parameter references filled	
	if (opts.noErrors()) {

	     // check for verbose argument
		_verbose = opts.argExist("v");
	    
	     // check output landscape topology type
	    if (opts.getStrVal("outType").compare("MS") == 0) {
	    	outType = MS;
	    } else if (opts.getStrVal("outType").compare("BT") == 0) {
	    	outType = BT;
	    } else if (opts.getStrVal("outType").compare("SN") == 0) {
	    	outType = SN;
	    }

	    if (LT != NULL) {
	    	delete LT;
	    }
		switch( outType ) {
		case MS :
			LT = new MinimaSet();
			break;
		case BT :
			LT = new BarrierTree();
			break;
		case SN :
			LT = new SaddleNet();
			break;
		default :
			std::cerr <<"\nERROR : given landscape topology type not handled !\n";
			return -2;
		}


		//! check for sequence and structure arguments
	    _sequence = opts.getStrVal("seq");
	    _structure= "";
	    
	      // check if structure is given, if not calculate mfe structure
	    if (opts.argExist("str")) {
	    	_structure = opts.getStrVal("str");
	    } else {
	    	char mfe_str[_sequence.size()];
	    	  // calculate mfe structure via Vienna RNA package
	    	fold(_sequence.c_str(), mfe_str);
	    	_structure = std::string(mfe_str);
	    }
	    
	    if(_sequence.size() != _structure.size()) {
	    	cerr << "\nERROR: sequence string and structure string must have equal length!\n";
	    	return false;
	    }
	    
	    if(!RNAFreeEnergy::getSequenceAlphabet()->isAlphabetString(_sequence)) {
	    	cerr << "\nERROR: given sequence string uses wrong alphabet!\n";
	    	return false;
	    }
	    
	    if(!RNAStructure::getStructureAlphabet()->isAlphabetString(_structure)) { 		    	
    		cerr << "\nERROR: given structure string uses wrong alphabet!\n";
	    	return false;
	    };
	    
	    RNAFreeEnergy* rnaFE = new RNAFreeEnergy(_sequence,_structure);
	    if( !rnaFE->isValid() ) {
		    	if( !rnaFE->hasValidStructure() )
		    		cerr <<"\nERROR: structure is not nested, number of ( has te be equal to number of )!\n";
		    	if( !rnaFE->hasMinLoopLength() )
		    		cerr <<"\nERROR: given structure has loops with size lower than "
		    				<<rnaFE->getMinLoopLength()<<"bases!\n";
	    		if( !rnaFE->hasValidBasePairs() )
	    			cerr <<"\nERROR: at least one base pairing is not allowed!(allowed: A-U, C-G and G-U)"<<endl;
	    	
    			delete rnaFE;
    			return false;
	    }   
	    delete rnaFE;
		    
    	  // input landscape topology given, that has to be read and extended
	    if (opts.argExist("inTP")) {
	    	std::istream * input = &std::cin;
	    	if (opts.getStrVal("inTP").compare("STDIN") != 0) {
	    		// filename given
	    		input = new std::ifstream(opts.getStrVal("inTP").c_str());
	    	}
	    	if (input->bad()) {
	    		std::cerr <<"\nERROR : cannot access input landscape from '"
	    					<<opts.getStrVal("inTP")
	    					<<"' !\n";
	    		return -1;
	    	}
			switch( outType ) {
			case MS :
				*input >>*dynamic_cast<MinimaSet*>(LT);
				break;
			case BT :
				*input >>*dynamic_cast<BarrierTree*>(LT);
				break;
			case SN :
				*input >>*dynamic_cast<SaddleNet*>(LT);
				break;
			default :
				std::cerr <<"\nERROR : given landscape topology type not handled !\n";
				return -2;
			}
	    	  // free memory
	    	if (opts.getStrVal("inTP").compare("STDIN") != 0) {
	    		(static_cast<std::ifstream*>(input))->close();
	    		delete input;
	    	}
	    	
	    }

	    //! check for iterations argument
	    if( opts.getIntVal("iter") < 0 ) {
	    	cerr << "\nERROR : iter has to be >= 0!\n";
	    	return false;
	    } else {  
	    	iterations = static_cast<size_t>(opts.getIntVal("iter"));
	    }
	    
    	//! check for optional outsteps argument	    
    	outSteps = 0;
	    if(opts.argExist("outEach")) {
	    	if(opts.getIntVal("outEach") < 0 ) {
	    		cerr << "\nERROR : out has to be >= 0!\n";
	    		return false;
	    	} else
    			outSteps = static_cast<size_t>(opts.getIntVal("outEach")) ;
	    }

	    //! check for optional randseed argument	    
	    if(opts.getIntVal("randSeed") < 0) {
	    		cerr << "\nERROR : random seed has to be >= 0!\n";
	    		return false;
	    }
	    seed = opts.getIntVal("randSeed");
	    
	    //! set optional energy treshold argument	    
	   	    Emax = opts.getDoubleVal("Emax");
	   	    
    	
	    //! check if iterations == 0 
	    if(iterations == 0) {
	    	if (opts.argExist("ncsteps") &&  opts.argExist("minMax")) {
	    		std::cerr <<"\nERROR : either ncsteps OR maximal minima number can be set !\n";
	    		return false;
	    	}
	    	
	    	//! abort when no new minimum found within last ncsteps
	    	if(opts.argExist("ncsteps")){
		    	if(opts.getIntVal("ncsteps") < 1 ) {
		    		cerr << "\nERROR : ncsteps has to be at least 1!\n";
		    		return false;
		    	} 
	    		noChangeSteps = static_cast<size_t>(opts.getIntVal("ncsteps")) ;
	    		if (outSteps == 0)  {
	    			_sac = new ell::SAC_MinChange(noChangeSteps);
	    		} else {
	    			_sac = new SAC_MinChange_Output(noChangeSteps, cout, outSteps);
	    		}
    			sac_type=1;
	    	} else if(opts.argExist("minMax")){    
	    	//! abort when certain amount of minima found
	    		if(opts.getIntVal("minMax") <= 0 ) {
	    			cerr << "\nERROR : maximal minima number has to be at least 1 !\n";
	    			return false;
	    		}else
	    		//! ugly hack to ensure benchmark bash script functionality
	    		minMax = (int)(opts.getIntVal("minMax"));
	    		if (outSteps=0) {
	    			_sac = new SAC_MaxMinima(minMax);
	    		} else {
	    			_sac = new SAC_MaxMinima_Output(minMax,cout,outSteps,0,0);
	    		}
    			sac_type=2;
	    	} else {
	    		std::cerr<<"\nERROR : iterations == 0 but neither ncsteps nor minMax given !\n";
	    		return false;
	    	}
	   }
	   else {
	    	if (opts.argExist("ncsteps")) {
	    		std::cerr <<"\nERROR : iterations limit is given : no ncsteps can be set !\n";
	    		return false;
	    	}
	    	if (opts.argExist("minMax")) {
	    		std::cerr <<"\nERROR : iterations limit is given : no maximal minima number can be set !\n";
	    		return false;
	    	}
	    	if (outSteps == 0) {
	    		_sac = new SAC_MaxCount(iterations);
	    	} else {
	    		_sac = new SAC_MaxCount_Output(iterations, cout, outSteps);
	    	}
	    	sac_type=3;
	   }
	    //! check for length argument
	    if( opts.getIntVal("len") < 0 ) {
	       	cerr << "\nERROR : len has to be at least 0!\n";
	       	return false;
	    } else {
	    	length = static_cast<size_t>(opts.getIntVal("len"));
	    }
	    	
	    switch(opts.getIntVal("ms")) {
	    	case 1:
	    		_ms = new MS_Successive();
	    		break;
	    	case 2:
	    		_ms = new MS_Uniform();
	    		break;
	    	case 3:
	    		if(opts.argExist("bf")) {
	    			boltz_factor = opts.getDoubleVal("bf");
	    			if(boltz_factor<1) {
	    				cerr << "\nERROR : bf has to be >=1 !\n";
	    				return false;
	    			}	    			
	    		}  else { 
	    			boltz_factor = 1;
	    		}
	    		_ms = new MS_Boltzmann(boltz_factor);	    			    			
	    		break;
	    	default: 
	    		//TODO: don't forget to refresh the error message when # of options changed
	    		cerr << "\nERROR : ms has to be element of {1,2,3}";
	    		return false;
	    } // switch(ms)
	    //! check for up argument
	    switch(opts.getIntVal("up")) {
	    	case 1:
	    	// fixed up walk length
	    		_wacup = new WAC_MaxLength(length);
	    		break;
	    	case 2:
	    		if(opts.argExist("lenMin")) {
	    			int walkUp_minLength_ = opts.getIntVal("lenMin");
	    			if( walkUp_minLength_ < 1 || walkUp_minLength_ > (int)length) {
	    				cerr << "\nERROR : lenMin has to be smaller or equal to len and at least 1 !\n";
	    				return false;
	    			} else 
	    				walkUp_minLength = static_cast<size_t>(walkUp_minLength_);
	    		} else {
	    			walkUp_minLength = 1;
	    		}
	    		
//	    		  // exponential distribution
//	    		if(opts.argExist("exp") && opts.getStrVal("exp").compare("1") != 0) {
//	    			walkUp_exp = opts.getDoubleVal("exp");
//	    			if(walkUp_exp < 0) {
//	    				cerr << "\nERROR : exp has to be at least 0!\n";
//	    				return false;
//	    			}
//	    			_wacup = new WAC_Random_Exp(walkUp_minLength, length, walkUp_exp);    		
//	    		} else {
	    		  // uniform distribution
	    			_wacup = new WAC_RandLength_UD(walkUp_minLength,length);
//	    		}
	    		break;
	    	default: 
	    		//TODO: don't forget to refresh the error message when # of options changed
	    		cerr << "\nERROR : up has to be element of {1,2}!\n";
	    		return false;
	    } // switch(up)
	    _up  = new WalkRandom(_wacup);
	    
	    //! check for down argument
	    switch(opts.getIntVal("down")) {
	    	case 1:
	    		_down = new WalkGradient(true);
	    		break;
	    	case 2:
	    		_down = new WalkAdaptive(WalkAdaptive::RANDOM, true);
	    		break;
	    	default:
	    		//TODO: don't forget to refresh the error message when # of options changed
	    		cerr << "\nERROR : down has to be element of {1,2}!\n";
	    		return false;
	    }// switch(down)
	    
	    //! check for outfile argument
	    _outfile = opts.getStrVal("outTP");
	    if (_outfile.size() == 0) {
	    	cerr <<"\nERROR : no file name etc. for output specified (outTP)!\n";
	    	return false;
	    }
	    
	    	//! init randomizer as an LCG for platform independance	
		RandomNumberGenerator* rng = new RNG_LCG();
		RNF::setRNG( rng );
		delete rng;
		rng=NULL;	
		RNF::getRNG().setSeed(seed);
	    	    	    
	}  // if (opts.noErrors())
	else 
		return false;
	
	
	///////////////////////  PARAMETER SETTINGS //////////////////////////////
	
	if (_verbose) {
		std::cout
			<<"\n RNA sequence    : " <<_sequence
			<<"\n RNA structure   : " <<_structure
			<<"\n LT output       : " <<_outfile
			<<"\n LT type         : " <<opts.getStrVal("outType")
			<<"\n UP min select   : " <<opts.getStrVal("ms")
			;
		if (opts.getIntVal("ms")==3) {
			std::cout 
			<<"\n UP min BF       : " <<boltz_factor;
		}
		std::cout
			<<"\n UP mode         : " <<opts.getStrVal("up")
			<<"\n UP length       : " <<length
			<<"\n UP length min   : " <<(opts.getIntVal("up")==2?walkUp_minLength:length)
			<<"\n DOWN mode       : " <<opts.getStrVal("down")
			<<"\n MIN max. E      : " <<(opts.argExist("Emax")?opts.getStrVal("Emax"):"---")
			;
		if (opts.argExist("ncsteps")) {
			std::cout 
			<<"\n ITER no change  : " <<opts.getIntVal("ncsteps");
		} else {
			std::cout 
			<<"\n ITER max        : " <<iterations;
		}
		std::cout
			<<std::endl
			<<std::endl;
	}
	
	//parsing successfull
	return true;
}

