#include "ell/MinimumSelector.hh"
#include <biu/RandomNumberFactory.hh>
#include <biu/assertbiu.hh>
#include <cmath>
#include <iostream>


namespace ell
{
////////// MinimumSelector SUCCESSIVE ////////////////
	MS_Successive::MS_Successive()
		: counter(0)	
	{ };
	
	size_t 
	MS_Successive::select(LandscapeTopology* pLT) const {
		assertbiu(pLT->getMinCount()!=0,"# of minima is 0 is given topology!");
		counter = ++counter % pLT->getMinCount(); 
		return counter;
	};

////////// MinimumSelector UNIFORM ////////////////////////
	MS_Uniform::MS_Uniform() 
	{};
	
	size_t 
	MS_Uniform::select(LandscapeTopology* pLT) const {
		//! generate random number in interval [0, n) where n is the number of minima
		return biu::RNF::getRN(pLT->getMinCount());
	};
	
////////// MinimumSelector BOLTZMANN/////////////////	
	MS_Boltzmann::MS_Boltzmann() 
		: lastLT(NULL), lastMinCount(0), boltzPrefixSums(), factor(1) 	
	{};
	
	MS_Boltzmann::MS_Boltzmann(double f) 
		: lastLT(NULL), lastMinCount(0), boltzPrefixSums(), factor(f) 	
	{};
	
	
	size_t 
	MS_Boltzmann::select(LandscapeTopology* pLT) const {
		double E, Elast, r;
		//! beta = 1 / R*T with R the gas constant 1.98717(cal/K*mol) and Temperature T = 37C 
		double beta = (double)1. / ((double)(273.15 + 37) * (double)1.98717 );
		

		//! if internal data doesn't belong to the given LT
		if(lastLT != pLT) {
			lastMinCount = 0;
			boltzPrefixSums.clear();
			lastLT = pLT;
		}

		Elast = (lastMinCount==0) ? 0 : boltzPrefixSums[lastMinCount-1];

		for (size_t i = lastMinCount; i < pLT->getMinCount(); ++i) {
			E = exp(-1. * beta * (pLT->getMin(i))->getEnergy() * factor);
//			std::cerr << "pLT->getMin(i))->getEnergy() ist "<< pLT->getMin(i)->getEnergy() <<std::endl;// TODO remove
//			std::cerr << "Elast ist "<< Elast <<std::endl;// TODO remove
//			std::cerr << "E ist "<< E <<std::endl;// TODO remove
			
			boltzPrefixSums.push_back(E+Elast);
			Elast += E;
		}

		lastMinCount = pLT->getMinCount();
		
		//! generate random number r in interval [0, Elast)
		r = (((double) biu::RNF::getRN()) / ((double) biu::RNF::getMaxRN())) * Elast;
				
		//! binary search on the prefixSums vector with random number r
		int lowerbound = 0, upperbound = boltzPrefixSums.size()-1;
		size_t position = (lowerbound + upperbound) / 2;
		
		while(lowerbound < upperbound) {
			if(boltzPrefixSums[position] > r)
				upperbound = position;
			else                          
				lowerbound = position + 1;
			position = (lowerbound + upperbound) / 2;
		}
//		std::cerr << "beta ist "<< beta <<std::endl;// TODO remove		
//		std::cerr << "r ist "<< r <<std::endl;
//		std::cerr << "Elast ist "<< Elast <<std::endl;
//		if(position==2) std::cerr << "MS_Bolztmann::select() returned MH"<<std::endl;// TODO remove
//		else std::cerr << "MS_Bolztmann::select() returned "<< position <<std::endl;// TODO remove
		return position;
	};

		
} // namespace ell
