#ifndef MINIMUMSELECTOR_HH_
#define MINIMUMSELECTOR_HH_


#include "ell/tp/LandscapeTopology.hh"

namespace ell
{
	/*!
	 * This abstract class is used to select the next minimum in which to start the sampling
	 * based on data from the LandScapeTopology
	 * 
	 * @author Hannes Kochni, Daniel Maticzka
	 */
	class MinimumSelector
	{
	public:
		
		virtual ~MinimumSelector(){};
		
		/*!
		 * Returns true if the StateCollector pLT fulfills a defined property.
		 */
		virtual size_t select(LandscapeTopology* pLT) const = 0;
	};
	
	
	/*!
	 * This subclass of MinimumSelector selects a minimum 
	 * based on a discrete uniform probability distribution.
	 *  
	 * @author Hannes Kochni, Daniel Maticzka
	 */
	class MS_Uniform : public MinimumSelector
	{
	protected:
		
	public:
		/*!
		 * Constructs a MinimumSelector that selects a minimum
		 * based on uniform probability of all minima
		 */
		MS_Uniform();
		virtual size_t select(LandscapeTopology* pLT) const;
	};
	
	/*!
	 * This subclass of MinimumSelector selects a minimum 
	 * based on a discrete Boltzmann probability distribution
	 * which is based on the energy of the minima.
	 *  
	 * @author Hannes Kochni, Daniel Maticzka
	 */
	class MS_Boltzmann : public MinimumSelector
	{
	protected:
		//! pointer to the last used LandscapeTopology
		mutable LandscapeTopology*	lastLT;
		//! last number of minima to check for new minima
		mutable	size_t 				lastMinCount;
		//! prefix sums of boltzmann weights for computing the probabilities
		mutable	std::vector<double>	boltzPrefixSums;
		//! a factor for all energies(==quotient for the temperature)
		//! to adjust the likeliness of choosing a low minimum
		//! factor > 1 means minima with low energy are more likely chosen than
		//! with normal Boltzmann distribution
		const 	double 				factor;
		
	public:
		/*!
		 * Constructs a MinimumSelector that selects a minimum
		 * based on a discrete Boltzmann probability distribution
		 * which is based on the energy of the minima.
		 */
		MS_Boltzmann();
		MS_Boltzmann(double factor);
		virtual size_t select(LandscapeTopology* pLT) const;
	};
	
	
	/*!
	 * This subclass of MinimumSelector selects a minimum 
	 * based on a counter which just increases with every select().
	 * Because indexing of new minima is in successive order, each
	 * minimum is selected one time.
	 *  
	 * @author Hannes Kochni, Daniel Maticzka
	 */
	class MS_Successive : public MinimumSelector
	{
	protected:
		mutable size_t counter;
		
	public:
		/*!
		 * Constructs a MinimumSelector that selects a minimum
		 * based on a counter which just increases with every select().
		 */
		MS_Successive();
		virtual size_t select(LandscapeTopology* pLT) const;
	};
		
} // namespace ell

#endif /*MINIMUMSELECTOR_HH_*/
