#include <biu/assertbiu.hh>
#include "ell/tp/SaddleNet.hh"
#include "ell/tp/BarrierTree.hh"
#include "ell/SamplingAbortionCriterion.hh"
#include <iostream>

using namespace std;

namespace ell {
SamplingAbortionCriterion::~SamplingAbortionCriterion() {
}
;

SAC_Output::SAC_Output(std::ostream& out, size_t outI) :
	ostr(&out), outInterval(outI), outCount1(0), outCount2(0) {
}

SAC_Output::~SAC_Output() {
}
;

void SAC_Output::output(LandscapeTopology* pLT, size_t val, bool enforce) const {
	outCount1 += val;
	if ( (outInterval!=0 && outCount1%outInterval==0) || enforce) {

		if (outInterval!=0)
			outCount2 = outCount1/outInterval;

		SaddleNet* pSN = dynamic_cast<SaddleNet*>(pLT);
		BarrierTree* pBT = dynamic_cast<BarrierTree*>(pLT);
		//! if first output write a header..
		/*if (outCount1== outInterval) {
		 *ostr << endl;
		 *ostr << "steps" <<"\t" << "minima" <<"\t";
		 if (pSN != NULL)
		 *ostr << "saddles" <<"\t" << "med. saddle" <<"\t";
		 else if (pBT != NULL)
		 *ostr << "medium barrier" <<"\t";
		 }*/

		*ostr << endl;
		//! write: walkindex, number of minima every time
		*ostr << outCount1 <<"\t" << pLT->getMinCount() <<"\t";
		if (pSN != NULL) {
			*ostr << pSN->getSaddleCount() <<"\t" << pSN->getMediumSaddleVal();
		} else if (pBT != NULL) {
			*ostr << pBT->getMediumBarrier();
		}
	}
}
void SAC_Output::setStepNum(size_t steps) {
	outCount1 = steps;
}
size_t SAC_Output::getStepNum() {
	return outCount1;
}
SAC_MaxCount::SAC_MaxCount(const size_t max) :
	maxCount(max), count(0) {
}

bool SAC_MaxCount::abort(LandscapeTopology* pLT, size_t val) const {
	count = count + val;
	return (count > maxCount);
}
;

SAC_MaxCount_Output::SAC_MaxCount_Output(const size_t max, std::ostream& out,
		const size_t outInt) :
	SAC_MaxCount(max), SAC_Output(out, outInt) {
}
;

bool SAC_MaxCount_Output::abort(LandscapeTopology* pLT, size_t val) const {
	//! increase count by value
	count += val;

	bool result = (count > maxCount);

	output(pLT, val, result);

	return result;
}
;

SAC_Delta::SAC_Delta(LandscapeTopology* p, const double d, const size_t testInt) :
	delta(d), pLT2(p), testInterval(testInt) {
	counter = 1;
	assertbiu(testInterval>0, "testInterval <= 0!");
}
;

bool SAC_Delta::abort(LandscapeTopology* pLT, size_t val) const {
	++counter;

	return (counter%testInterval == 0) ? (pLT2->getRMSD(pLT) <= delta) : false;
}
;
SAC_Delta_Output::SAC_Delta_Output(LandscapeTopology* pLT2, const double delta,
		const size_t testInterval, std::ostream& ostr, const size_t outInterval) :
	SAC_Delta(pLT2, delta, testInterval), SAC_Output(ostr, outInterval) {
}
;

bool SAC_Delta_Output::abort(LandscapeTopology* pLT, size_t val) const {
	bool result = (counter%testInterval == 0) ? (pLT2->getRMSD(pLT) <= delta)
			: false;
	output(pLT, val, result);
	if (counter%outInterval == 0)
		*ostr <<"\t\t" <<pLT2->getRMSD(pLT);

	++counter;

	return result;
}
;

SAC_MinChange::SAC_MinChange(size_t noCha) :
	noChange(noCha), minCount(0), noChangeCount(0) {
}
;

bool SAC_MinChange::abort(LandscapeTopology* pLT, size_t val) const {
	if (minCount != pLT->getMinCount()) {
		minCount = pLT->getMinCount();
		noChangeCount = 0;
	} else
		noChangeCount += val;

	return (noChangeCount > noChange);
}
;

SAC_MinChange_Output::SAC_MinChange_Output(const size_t noCha,
		std::ostream& out, const size_t outInt) :
	SAC_MinChange(noCha), SAC_Output(out, outInt) {
}
;

bool SAC_MinChange_Output::abort(LandscapeTopology* pLT, size_t val) const {

	if (minCount != pLT->getMinCount()) {
		minCount = pLT->getMinCount();
		noChangeCount = 0;
	} else
		noChangeCount += val;

	bool result = (noChangeCount > noChange);

	output(pLT, val, result);

	return result;
}
;

SAC_MaxMinima::SAC_MaxMinima(const size_t max) :
	maxMinima(max) {
}
bool SAC_MaxMinima::abort(LandscapeTopology* pLT, size_t val) const {
	return (pLT->getMinCount() >= maxMinima);
}
;

SAC_MaxMinima_Output::SAC_MaxMinima_Output(const size_t max, std::ostream& out,
		const size_t outInt, const size_t lmincount, const size_t lcount) :
	SAC_MaxMinima(max), SAC_Output(out, outInt), no_min_found(0) {
	last_min_count = lmincount;
	count = lcount;
}
;

bool SAC_MaxMinima_Output::abort(LandscapeTopology* pLT, size_t val) const {
	//!compute how much steps are necessary to discover a new minimum
	ratio=((double)count/(double)pLT->getMinCount());
	//!whenever a new minimum is found print output	
	if (last_min_count < pLT->getMinCount() ) {
		*ostr<<(double)(pLT->getMinCount())/(double)(maxMinima)*100<<"\t\t"
				<<ratio<<"\t\t"<<no_min_found<<endl;
		no_min_found=0;
		last_min_count = pLT->getMinCount();
	}
	count+=val;
	no_min_found++;
	return (pLT->getMinCount() >= maxMinima);
}
;
double SAC_MaxMinima_Output::getCount() {
	return count;
}
;

/*!
 * This subclass of SamplingAbortionCriterion never aborts the sampling.
 * 
 * @author Hannes Kochni�, Daniel Maticzka
 */
SAC_OpenEnd::SAC_OpenEnd() {
}
;

bool SAC_OpenEnd::abort(LandscapeTopology* pLT, size_t val) const {
	return false;
}
;

SAC_DeltaHeight_Output::SAC_DeltaHeight_Output(std::ostream& out, size_t outI,
		size_t noChaHe, size_t testInt) :
	SAC_Output(out, outI), heightNoChange(noChaHe), currBarrierHeight(65536),
			tmpnc(0), testInterval(testInt) {
}

bool SAC_DeltaHeight_Output::abort(LandscapeTopology* pLT, size_t val) const {
	/*!calculation of mean barrier height is expensive 
	 * so calculate only every testInterval steps
	 */
	bool result;
	if (counter%testInterval == 0) {
		BarrierTree* pBT = dynamic_cast<BarrierTree*>(pLT);
		SaddleNet* pSN = dynamic_cast<SaddleNet*>(pLT);
		if (pBT != NULL) {
			if (pBT->getMediumBarrier() < currBarrierHeight) {
				currBarrierHeight = pBT->getMediumBarrier();
				tmpnc=0;
			} else
				tmpnc++;
		} else if (pSN != NULL) {
			if (pSN->getMediumSaddleVal() < currBarrierHeight) {
				currBarrierHeight = pSN->getMediumSaddleVal();
				tmpnc=0;
			} else
				tmpnc++;
		}
		result = (tmpnc > heightNoChange);
	} else
		result=false;
	output(pLT, val, result);
	++counter;
	return result;
}
;
SAC_DeltaNoChange_Output::SAC_DeltaNoChange_Output(LandscapeTopology* pLT2,
		double delta, size_t testInt, std::ostream& out, size_t outI,
		size_t noCha) :
			SAC_Delta(pLT2, delta, testInt), SAC_Output(out, outI), currRMSD(INT_MAX),
			noChaRMSD(0), noChange(noCha) {
}
bool SAC_DeltaNoChange_Output::abort(LandscapeTopology* pLT, size_t val) const {
	//cerr<<"noChange:\t"<<noChange<<"\tnoChaRMSD:\t"<<noChaRMSD<<endl;
	cerr<<"counter: "<<counter<<"testInterval: "<<testInterval<<endl;
	if (counter%testInterval == 0) {
		if (pLT2->getRMSD(pLT) == currRMSD)
			noChaRMSD+=val;
		else {
			currRMSD = pLT2->getRMSD(pLT);
			noChaRMSD=0;
		}
	}
	output(pLT, val, false);
	if (counter%outInterval == 0)
		*ostr <<"\t\t" <<pLT2->getRMSD(pLT);
	++counter;
	return (noChange < noChaRMSD || (currRMSD < delta));
}
;
void SAC_DeltaNoChange_Output::resetCounter() {
	noChaRMSD=0;
}
size_t SAC_DeltaNoChange_Output::getCounter() {
	return counter;
}
} // namespace ell
