#ifndef SAMPLINGABORTIONCRITERION_HH_
#define SAMPLINGABORTIONCRITERION_HH_

#include "ell/tp/SaddleNet.hh"
#include "ell/tp/BarrierTree.hh"
#include <iostream>

namespace ell {
/*!
 * This abstract class is used to check if a sampling algorithm should finish
 * based on data from the LandScapeTopology
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SamplingAbortionCriterion {
public:

	virtual ~SamplingAbortionCriterion();

	/*!
	 * Returns true if the LandscapeTopology pLT fulfills a defined property.
	 * In certain cases val is also used (e.g. SAC_MaxCount) 
	 */

	virtual bool abort(LandscapeTopology* pLT, size_t val = 1) const = 0;
};

class SAC_Output {
protected:
	mutable std::ostream* ostr;
	//! number counts between output of minimacount and medium saddleHeight
	const size_t outInterval;
	//! counter for checking if there should be an output
	mutable size_t outCount1;
	//! second counter for checking if there should be an output
	mutable size_t outCount2;

public:
	SAC_Output(std::ostream&, size_t outInterval);
	virtual ~SAC_Output();

	/*!
	 * Prints an output every outInterval steps, depending on type of given LandscapeTopology
	 * @param pLT the topology to get the information from
	 * @param enforce if true the output will be streamed to ostr even when not outInterval steps have been passed
	 * @param val the # of steps since the last invocation of the output function
	 */
	void output(LandscapeTopology* pLT, size_t val, bool enforce = false) const;
	void setStepNum(size_t steps);
	size_t getStepNum();
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if 
 * a counter reaches a maximum. Normally this counter equals the number of walks 
 * done, but this is up to the user of this class.
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_MaxCount : public SamplingAbortionCriterion {
protected:
	//! const maximum number of counts till abortion
	const size_t maxCount;
	//! counter for AbortionCriterion, mutable because
	//! it has to be changed in const function abort()   
	mutable size_t count;

public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts when a maximum
	 * number of walks is reached.
	 * @param max maximum number of walks
	 */
	SAC_MaxCount(const size_t max);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if 
 * a counter reaches a maximum. Normally this counter equals the number of walks 
 * done, but this is up to the user of this class.
 * It also outputs the number of minima in the topology as well as the
 * medium saddle height every outInterval counts. 
 * If outInterval=0 there is only one output when abort returns true.
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_MaxCount_Output : public SAC_MaxCount, public SAC_Output {
public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts when a maximum
	 * number of walks is reached.
	 * @param max maximum number of walks
	 */
	SAC_MaxCount_Output(const size_t max, std::ostream& ostr,
			const size_t outInterval = 0);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
	//!dummy method for resetting count
	//void resetCount();
	size_t getCount();
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * RMSD between 2 SaddleNets reachs delta 
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_Delta : public SamplingAbortionCriterion {
protected:
	//! the RMSD between pLT and pLT2 at which to abort
	const double delta;
	//! pointer to flooded (second) LT
	LandscapeTopology* pLT2;
	//! rmsd gets only computed each testInterval steps  
	const size_t testInterval;
	//! counter for checking when testInterval steps are reached   
	mutable size_t counter;

public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * RMSD between the SaddleNets LT and LT2 is <=delta
	 */
	SAC_Delta(LandscapeTopology* pLT2, const double delta,
			const size_t testInterval);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;

};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * RMSD between 2 SaddleNets reachs delta, and writes the #of minima, saddle
 * and average saddle height to a given outstream every outInterval steps  
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_Delta_Output : public SAC_Delta, public SAC_Output {
public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * number of minima didn't change for NoChange steps.
	 */
	SAC_Delta_Output(LandscapeTopology* pLT2, const double delta,
			const size_t testInterval, std::ostream& ostr,
			const size_t outInterval);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * minima didn't change for xx steps
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_MinChange : public SamplingAbortionCriterion {
protected:
	//! number of steps after which to abort if # of minima didn't change
	const size_t noChange;
	//! variable to track the last number of minima
	mutable size_t minCount;
	//! counter for # of steps since last change of minimanumber
	mutable size_t noChangeCount;
public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * number of minima didn't change for NoChange steps.
	 */
	SAC_MinChange(const size_t noChange);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * minima didn't change for xx steps.
 * 
 * 
 */
class SAC_MinChange_Output : public SAC_MinChange, public SAC_Output {
public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * number of minima didn't change for NoChange steps.
	 */
	SAC_MinChange_Output(const size_t max, std::ostream& ostr,
			const size_t outInterval = 0);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * number of found minima exceeds a certain limit.
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_MaxMinima : public SamplingAbortionCriterion {
protected:
	//! energy boundary that won't be exceeded
	const size_t maxMinima;

public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * number of found minima exceeds a certain limit.
	 * @param max number of minima after which to abort
	 */
	SAC_MaxMinima(const size_t max);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * number of found minima exceeds a certain limit.
 * Whenever a minmum is found it prints:
 * % Minima (compared to maxMinima) found so far
 * Ratio: #minima/#steps
 */
class SAC_MaxMinima_Output : public SAC_MaxMinima, public SAC_Output {
protected:
	//!average ratio steps/found minima
	mutable double ratio;
	//! variable to keep track of new found minima
	mutable size_t last_min_count;
	//! a counter
	mutable size_t count;
	mutable size_t no_min_found;

public:
	/*!
	 * Constructs a SamplingAbortionCriterion that aborts the sampling if the
	 * number of found minima exceeds a certain limit.
	 * @param max number of minima after which to abort
	 */
	SAC_MaxMinima_Output(const size_t max, std::ostream& ostr,
			const size_t outInterval = 0, const size_t last_min_count = 0,
			const size_t count =0);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
	double getCount();
};

/*!
 * This subclass of SamplingAbortionCriterion never aborts the sampling.
 * 
 * @author Hannes Kochniss, Daniel Maticzka
 */
class SAC_OpenEnd : public SamplingAbortionCriterion {
public:
	SAC_OpenEnd();
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};

/*!This subclass of SamplingAbortionCriterion aborts sampling if  
 * medium saddle height or medium barrier height  did not  decrease
 *  vor xx steps.
 * TODO: evtl von SAC_Delta ableiten
 */

class SAC_DeltaHeight_Output : public SAC_Output,
	public SamplingAbortionCriterion {
private:
	//!holds how long the medium barrier/saddle height should not change before sampling should abort
	mutable size_t heightNoChange;
	//! holds current barrier/saddle height
	mutable double currBarrierHeight;
	mutable size_t tmpnc;
	const size_t testInterval;
	mutable size_t counter;
public:
	SAC_DeltaHeight_Output(std::ostream&, size_t outInterval, size_t noChaHe,
			size_t testInterval);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
};
/*!
 * This subclass of SamplingAbortionCriterion aborts the sampling if the
 * RMSD between 2 SaddleNets does not change for xx steps or drops below a given value and prints results
 * TODO: ausfuehrlicher
 */
class SAC_DeltaNoChange_Output : public SAC_Delta, public SAC_Output {
public:
	SAC_DeltaNoChange_Output(LandscapeTopology* pLT2, const double delta,
			const size_t testInterval, std::ostream& ostr,
			const size_t outInterval, const size_t noChange);
	bool abort(LandscapeTopology* pLT, size_t val = 1) const;
	void resetCounter();
	size_t getCounter();
private:
	mutable double currRMSD;
	mutable size_t noChaRMSD;
	size_t noChange;
};
}
// namespace ell

#endif /*SAMPLINGABORTIONCRITERION_HH_*/
