#ifndef TOPOLOGYFLOODER_HH_
#define TOPOLOGYFLOODER_HH_

#include <vector>
#include <biu/Matrix.hh>
#include "ell/tp/LandscapeTopology.hh"
#include "ell/StatePairCollector.hh"
#include "ell/SPC_Counting.hh"

namespace ell {

	//! wrapper class for the floodAll(..) function
	//! has a static floodAll function and a member function floodAll
template <class SPC> 
class TopologyFlooder {
public:
	//! constructor
	TopologyFlooder (
					LandscapeTopology* pLT,
					std::vector<StateCollector*>& vSC, 
					biu::Matrix<StatePairCollector*>& mSPC,
					StateCollector* rejectedSC = NULL);

	virtual ~TopologyFlooder();
	
	  //! Fills succesivelly all basins of the given local minima in the 
	  //! landscape and stores the seen States in energy increasing order to
	  //! the given StateCollectors (given by a vector).
	  //! The basin is flooded up to maxE.
	  //! Therefore, all neighbored States with energy below maxE have to be 
	  //! stored in a queue. A maximal number of states to store
	  //! for investigation in parallel can be given by maxToStore. If the 
	  //! underlying queue size exeeds this maxE is decreased for deltaE and 
	  //! all States above the new maxE are removed from the queue. This is 
	  //! repeated until the queue size falls back below maxToStore.
	  //!
	  //! @param vMinima are all the local minima that are known
	  //! @param vSC is a StateCollector type that the accepted and checked states 
	  //!           are written to
	  //! @param maxE the maximal energy all states should be below
	  //! @param deltaE the delta E to lower maxE in case too many states are 
	  //!               in the queue
	  //! @param maxToStore the maximal number of elements the underlying queue
	  //!                   is allowed to contain
	  //! @return the maximal energy until the basin was flooded (all accepted 
	  //!         state have energy below)
	
	static 
	double
	floodAll(	LandscapeTopology* pLT, 
				std::vector<StateCollector*>& vSC, 
				biu::Matrix<StatePairCollector*>& mSPC,
				const double maxE, 
				const double deltaE, 
				const size_t maxToStore,
				StateCollector* rejectedSC = NULL
			);
	
	virtual
	double
	floodAll(	const double maxE,	
				const double deltaE,
				const size_t maxToStore
			);

private:
	LandscapeTopology* pLT; 
	std::vector<StateCollector*> vSC;
	biu::Matrix<StatePairCollector*> mSPC;
	StateCollector* rejectedSC;
};

}  // namespace ell

#include "ell/TopologyFlooder.icc"
#endif /*TOPOLOGYFLOODER_HH_*/
