#ifndef WALKABORTIONCRITERION__HH_
#define WALKABORTIONCRITERION__HH_

#include <ell/WalkAbortionCriterion.hh>
#include <biu/RandomNumberFactory.hh>
#include <cmath>

namespace ell {

/*!
 * This subclass of WalkAbortionCriterion aborts 
 * walks with lengths between and including minLength and maxLength
 * on a random basis 
 * WAC_Random_Exp(0,2,1) picks walklengths 0,1 and 2 with
 * even probability among the 3 choices (cause exponent is 1)
 * so it even aborts walks with length 0 -> no walk happens
 * the probability function for the interval [minLength,maxLength]
 * rises by ((length-minLength)/(maxlength-minlength))^exponent
 *  
 * @author Hannes Kochniss
 */
class WAC_Random_Exp : public WalkAbortionCriterion {
protected:
	//! minimum length of walk until abortion
	const size_t minLength;
	//! maximum length of walk until abortion
	const size_t maxLength;
	//! the exponent for the probability distribution
	//! between minLength and maxLength
	//! if exponent = 1 it's a linear distribution
	const double exponent;

public:
	/*!
	 * Constructs a WalkAbortionCriterion that aborts with a certain
	 * probability if length of walk >= minLength, and aborts for sure
	 * when length of walk reaches maxLength
	 * @param minL minimal length of walk
	 * @param maxL maximal length of walk
	 * @param exponent exponent used for the probability function between 
	 * minimal and maximal length of the Walk, 
	 * exponent <1: more probable that walk aborts near minLength
	 * exponent =1: abortion is euqally probable for all lengths in [minLength,maxLength]
	 * exponent >1: more probable that walk aborts near maxLength
	 */
	WAC_Random_Exp(const size_t minL, const size_t maxL, const double exponent);
	virtual ~WAC_Random_Exp() {
	}
	;
	bool abort(StateCollector* sc) const;
};

/*!
 * This subclass of WAC_Random_Exp works 
 * like WAC_Random_Exp except it uses exponent = 1
 *  
 * @author Hannes Kochniss
 */
class WAC_Random_Linear : public WAC_Random_Exp {

public:
	/*!
	 * see WAC_Random_Exp constructor..
	 * exponent is set to 1, so the probability function
	 * between minLength and maxLength is linear
	 * @param minL minimal length of walk
	 * @param maxL maximal length of walk
	 */
	WAC_Random_Linear(const size_t minL, const size_t maxL);
	virtual ~WAC_Random_Linear() {
	}
	;

};
/*!
 * Works like WAC_MaxLength but has setter/getter Methods 
 *  
 */
class WAC_MaxLengthSG : public WalkAbortionCriterion {
private:
	//! maximum number of accepted states
	size_t maxLength;

public:
	/*!
	 * Constructs a WalkAbortionCriterion that aborts when a maximum
	 * number of accepted states is reached.
	 * @param max maximum number of accepted states
	 */
	WAC_MaxLengthSG(const size_t max);
	//!length setter and getter methods
	bool setMaxLength(const size_t max);
	size_t getMaxLength();
	bool abort(StateCollector* sc) const;

};


/*!
 * This subclass of WalkAbortionCriterion aborts 
 * walks with random lengths in the interval [minLength,maxLength].
 * 
 * WAC_RandLength_UD(1,3) picks walklengths 1, 2, and 3 with
 * uniform probability among the 3 choices.
 * 
 * The probability function for the interval [minLength,maxLength]
 * follows a uniform distribution.
 * 
 * To minimize the number of calculations, the currently accepted maximal walk
 * length is calculated only during construction and everytime a walk was 
 * aborted. This allows for a reuse of the WAC after a walk was aborted such 
 * that the next walk length is chosen randomly as well. 
 *  
 * @author Martin Mann
 * @author Hannes Kochniss
 */
class WAC_RandLength_UD : public WalkAbortionCriterion {
protected:
	//! Minimal length of a walk until abortion.
	const size_t minLength;
	//! Maximal length of a walk until abortion.
	const size_t maxLength;
	//! The currently allowed maximal walk length. It is set during construction
	//! and changed to a random number in the interval [minLength,maxLength] 
	//! everytime a walk is aborted.
	mutable size_t curMaxLength;

	//! Calculates the next maximal walk length accepted within the interval
	//! [minLength,maxLength].
	//! The probability function for the interval follows a uniform distribution.
	//! @return a random number in the interval.
	size_t getNewCurMaxLength() const;
	
public:
	/*!
	 * Constructs a WalkAbortionCriterion that aborts with a certain
	 * probability if length of walk >= minLength, and aborts for sure
	 * when length of walk reaches maxLength
	 * @param minL minimal length of walk
	 * @param maxL maximal length of walk
	 */
	WAC_RandLength_UD(const size_t minL, const size_t maxL);
	virtual ~WAC_RandLength_UD();
	bool abort(StateCollector* sc) const;
};


} // namespace ell


#endif /*WALKABORTIONCRITERION_HH_*/
