#include "ell/directWalk.hh"
#include <biu/assertbiu.hh>

using std::string;
using std::vector;
using std::list;
using std::cerr;
using std::endl;
using namespace biu;

namespace ell
{

double addBond(string sequence, vector<size_t>& v_struct, size_t start, size_t end, StateCollector* sc) {
	v_struct[start] = end;
	
	//! convert vector of size_t back to a string
	string structure(v_struct.size(),'.');
	
	for (size_t i = 0; i < v_struct.size(); ++i) {
		if(v_struct[i]!=0) {
			structure[i]='(';
			structure[v_struct[i]]=')';
		}
	}
	
	S_RNAfe_SingleM* rna = new S_RNAfe_SingleM(sequence,structure);
	double energy = rna->getEnergy();
	
	if(sc != NULL) sc->add(*rna);
	
	delete rna;
	return energy;
}

double deleteBond(string sequence, vector<size_t>& v_struct, size_t start, StateCollector* sc) {
	v_struct[start] = 0;
	
	//! convert vector of size_t back to a string
	string structure(v_struct.size(),'.');
	
	for (size_t i = 0; i < v_struct.size(); ++i) {
		if(v_struct[i]!=0) {
			structure[i]='(';
			structure[v_struct[i]]=')';
		}
	}
	
	S_RNAfe_SingleM* rna = new S_RNAfe_SingleM(sequence,structure);
	double energy = rna->getEnergy();
	
	if(sc != NULL) sc->add(*rna);
	
	delete rna;
	return energy;
}

double directWalk( const State* _s1_, const State* _s2_, StateCollector* _sc_, const size_t _iterations_) {
	
	typedef vector<size_t>::iterator vit;
	
//! test if states are really RNA states and collector != NULL
	assertbiu(_s1_ != NULL, "State A is NULL");
	assertbiu(_s1_ != NULL, "State B is NULL");
	assertbiu(_sc_ != NULL, "StateCollector is NULL");
	
	const S_RNAfe_SingleM* rna_A = dynamic_cast<const S_RNAfe_SingleM*>(_s1_);
	const S_RNAfe_SingleM* rna_B = dynamic_cast<const S_RNAfe_SingleM*>(_s2_);
	
	assertbiu(rna_A != NULL, "state A is no RNA");
	assertbiu(rna_B != NULL, "state B is no RNA");

//! test if states have same sequence
	size_t length = rna_A->toString().size();
	assertbiu(length == rna_B->toString().size(), "state A and B have different sequence");
	assertbiu(rna_A->toString().substr(length/2,(length-2)/2) == rna_B->toString().substr(length/2,(length-2)/2), "rna A and B have different sequences");
	
//! get sequence and structure from RNA states, str_A str_B
	string sequence = rna_A->toString().substr(length/2,(length-2)/2);
	string str_A = rna_A->toString().substr(0, (length-2)/2 );
	string str_B = rna_B->toString().substr(0, (length-2)/2 );
	length = str_A.size();
	
	vector<size_t> structA, structB;
	vector<size_t> tempA, tempB;
		
//! get only differing state pairs as a list for A and B
	for (size_t i = 0; i < length; ++i) {
		structA.push_back(0);
		structB.push_back(0);
		
		if(str_A[i]=='(') tempA.push_back(i);
		if(str_A[i]==')') {
			structA[tempA.back()]=i;
			tempA.pop_back();
		}
				
		if(str_B[i]=='(') tempB.push_back(i);
		if(str_B[i]==')') {
			structB[tempB.back()]=i;
			tempB.pop_back();
		}
	}
	
	vector<size_t> indexA, indexB;
	
	for (size_t i = 0; i < length; ++i) {
		if(structA[i] != structB[i]) {
			if(structA[i] != 0) indexA.push_back(i);
			if(structB[i] != 0) indexB.push_back(i);
		}
	}

	
//! get conflicting pairs info for each pair in list B
	vector<vector<size_t> > conflicts;
	vector<size_t> noconflicts;
	vector<size_t> temp;
	vector<size_t> temp2(indexA.size(),0);
		
	for (size_t i = 0; i < indexB.size(); ++i) {
		temp.clear();
		for (size_t j = 0; j < indexA.size(); ++j) {
			if( !((indexB[i] < indexA[j] &&  structB[indexB[i]] > structA[indexA[j]]) ||
					(indexB[i] > indexA[j] &&  structB[indexB[i]] < structA[indexA[j]]))) {
				temp.push_back(j);
				temp2[j] = 1;
			}
		}		
		conflicts.push_back(temp);
	}
	
//! store the rest of the bonds of A (aka the nonconflicting pairs)
	for (size_t i = 0; i < indexA.size(); ++i) {
		if(temp2[i]==0) noconflicts.push_back(i); 
	}
	
	//! all choices of bonds in B with smallest conflicts
	vector<size_t> choices;
	//! the indices of bonds added, and non conflicting bonds deleted 
	vector<size_t> addedBonds, deletedNonConflBonds;
	//! copy of structure vector of sequence A
	vector<size_t> temp_struct;
	vector<size_t> temp_struct_best;
	//! copy of conflicts matrix
	vector<vector<size_t> > temp_conflicts;
	//! copy of nonconflicting bonds
	vector<size_t> temp_noconflicts;
	//! minimal number of conflicts over all bonds in target structure B
	size_t min_conflicts;
	//! indices for choices
	size_t choice_add, choice_delete, toAdd, toDelete;
	//! StateCollectors which remember all states seen
	SC_Listing *scl = NULL, *scl_best = NULL;
	//! energy values
	double energy = 0, peakEnergy = -INT_MAX, Emax = INT_MAX;
	
	
//! iterate some times with random choices:
	for (size_t count = 0; count < _iterations_; ++count) {
//		cerr << "iteration "<<count<<endl;
		scl = new SC_Listing();
		temp_struct = structA;
		temp_conflicts = conflicts;
		addedBonds.clear();
		peakEnergy = -INT_MAX;
		
		//! get all the bonds with lowest conflicts from B
		while(addedBonds.size() != conflicts.size()) {
			min_conflicts = INT_MAX;
			//! get all the bonds in B with minimal conflicts
			for (size_t i = 0; i < temp_conflicts.size(); ++i) {
				//! only if we haven't added this bond..
				if(find(addedBonds.begin(),addedBonds.end(),i) != addedBonds.end()) continue;
//				if(count < 1) cerr << " bond "<<i<< " with size "<<temp_conflicts[i].size()<<endl;
				if(temp_conflicts[i].size() < min_conflicts) { choices.clear(); min_conflicts = temp_conflicts[i].size();} 
				if(temp_conflicts[i].size() == min_conflicts) choices.push_back(i);
			}
//			cerr << "choices to Add are:\t";
//			for (size_t i = 0; i < choices.size(); ++i) {
//				cerr << choices[i] << " ";				
//			}
//			cerr << endl;
			//! choose randomly a bond to add
			choice_add = (size_t)((double)(choices.size()) * 
						 (double)(RNF::getRN(100)) / (double)100);
			//! index of bond to add in IndexB
			toAdd = choices[choice_add];
//			if(count<1) {cerr << "toAdd: " << toAdd<<endl;}
			
			while(temp_conflicts[toAdd].size() != 0) {
				//! choose bonds to delete in a random order from conflicts list
				choice_delete = (size_t)((double)(temp_conflicts[toAdd].size()) * 
								(double)(RNF::getRN(100)) / (double)100);
				//! index of bond to delete
				toDelete = temp_conflicts[toAdd][choice_delete];
				//! deletion of bond
				energy = deleteBond(sequence, temp_struct, indexA[toDelete], scl);
				//! if energy found is above energy of best walk -> abort
				if(energy > Emax) break;
				if(energy > peakEnergy) peakEnergy = energy;
				
				//! delete all occurings of deleted bond in conflicts list
//				if(count<1) {cerr << "toDelete: " << toDelete<<endl;}
				for (size_t i = 0; i < temp_conflicts.size(); ++i) {
										
					for (size_t j = 0; j < temp_conflicts[i].size(); ++j) {
						if(toDelete == temp_conflicts[i][j]) { 
							temp_conflicts[i].erase(temp_conflicts[i].begin()+j);
						}
					}
//					if(count<1) {
//						cerr << "temp_conflicts["<<i<<"]:\t";
//						for (size_t k = 0; k < temp_conflicts[i].size(); ++k) {
//							cerr << temp_conflicts[i][k] << " ";
//						}
//						cerr << endl;
//					}
				}
			}

			if(energy > Emax) break;
						
			energy = addBond(sequence, temp_struct, indexB[toAdd], structB[indexB[toAdd]], scl);
			addedBonds.push_back(toAdd);
			
			if(energy > peakEnergy) peakEnergy = energy;
			if(energy > Emax) break;				
		}
		
		if(energy > Emax) { delete scl; continue;}

		temp_noconflicts = noconflicts;
		deletedNonConflBonds.clear();
		//! delete the bonds in A which are not conflicting with bonds in B
		while(temp_noconflicts.size() != 0) {
			choice_delete = (size_t)((double)(temp_noconflicts.size()) * 
							(double)(RNF::getRN(100)) / (double)100);
			toDelete = temp_noconflicts[choice_delete];
			energy = deleteBond(sequence, temp_struct, indexA[toDelete], scl);
			if(energy > Emax) break;
			if(energy > peakEnergy) peakEnergy = energy;
			temp_noconflicts.erase(temp_noconflicts.begin() + choice_delete);
		}
			
		if(energy > Emax) { delete scl; continue;}

		Emax = peakEnergy;
		delete scl_best;
		scl_best = scl;
		temp_struct_best = temp_struct;
	}
	
	list<State*> l = scl_best->getList();
	
	for(list<State*>::const_iterator cit = l.begin(); cit != l.end(); ++cit) {
		_sc_->add(**cit);
	}
	

	string str_temp(temp_struct_best.size(),'.');
		
	for (size_t i = 0; i < temp_struct_best.size(); ++i) {
		if(temp_struct_best[i]!=0) {
			str_temp[i]='(';
			str_temp[temp_struct_best[i]]=')';
		}
	}
	
//	cerr << "str_A:\t\t" << str_A<<endl;
//	cerr << "str_temp:\t" << str_temp<<endl;
//	cerr << "str_B:\t\t" << str_B<<endl;
	
	delete scl_best;
	
	return Emax;
}

}// ell
