#ifndef MINSAMPLING_HH_
#define MINSAMPLING_HH_

#include <vector>
#include <ell/Walk.hh>
#include <ell/SC_Counting.hh>
#include "ell/tp/LandscapeTopology.hh"
#include "ell/SC_Maximum.hh"
#include "ell/MinimumSelector.hh"
#include "ell/SamplingAbortionCriterion.hh"

namespace ell
{
	/*!
	 * Performs an iup and down walk and updates the LandscapeTopology. 
	 * 
	 * Starting from the minimum with the specified index, the provided up-Walk 
	 * is performed followed by the specified down-Walk. The latter leads to a 
	 * minimum. If the minimum is not known and addNewMin==true, the minimum is 
	 * added to pLT and the State with maximal energy during the up-Walk is 
	 * added as a saddle point to pLT. If the minimum is already present, an 
	 * update of the saddle point is done.
	 * 
	 * @param up The upwards walk to perform, starting from a choosen minimum.
	 * @param down The downwards walk to apply to reach a minimum.
	 * @param pLT The landscape topology the minima are choosen from and added to.
	 * @param minToStartIndex The index of the minimum from pLT to start the 
	 *                        up-Walk from.
	 * @param Emax The maximally allowed energy for minima that are added to pLT.
	 * @param addNewMin Whether or not new minima should be added or only a 
	 *                  barrier/saddle refinement should be done.
	 * @return Whether or not the LandscapeTopology pLT was changed.
	 */
	bool 
	singleMinSampling(	const Walk*	const up,
						const Walk* const down,
						LandscapeTopology* const pLT,
						const size_t minToStartIndex, 
						const double Emax,
						const bool addNewMin = true);

/*	bool 
	singleMinSampling(	const Walk*	const up,
						const Walk* const down,
						LandscapeTopology* const pLT,
						const size_t startIndex,
						const size_t walkCount);
	*/
	
	/*!
	 * Performs an iteration of up and down walks. Each time a minimum of the 
	 * given LandscapeTopology is choosen via the provided MinimumSelector.
	 * Starting from that minimum the provided up-Walk is performed followed by
	 * the specified down-Walk. The latter leads to a minimum. If the minimum is
	 * not known and addNewMin==true, the minimum is added to pLT and the State
	 * with maximal energy during the up-Walk is added as a saddle point to pLT.
	 * If the minimum is already present, an update of the saddle point is done.
	 * 
	 * @param up The upwards walk to perform, starting from a choosen minimum.
	 * @param down The downwards walk to apply to reach a minimum.
	 * @param pLT The landscape topology the minima are choosen from and added to.
	 * @param ms The minimum selector used to select a minimum from pLT for the
	 *           next iteration.
	 * @param sac The sampling abortion criterion that stops the iteration.
	 * @param Emax The maximally allowed energy for minima that are added to pLT.
	 * @param addNewMin Whether or not new minima should be added or only a 
	 *                  barrier/saddle refinement should be done.
	 * @return Whether or not the LandscapeTopology pLT was changed.
	 */
	bool 
	allMinSampling(		const Walk*	const up,
						const Walk* const down,
						LandscapeTopology* const pLT,
						MinimumSelector*	ms,
						SamplingAbortionCriterion*	sac,
						const double Emax,
						const bool addNewMin = true);
	
}

#endif /*MINSAMPLING_HH_*/
