#ifndef BARRIERTREE_HH_
#define BARRIERTREE_HH_

#include "ell/tp/LandscapeTopology.hh"
#include "ell/tp/BarrierTree_IO.hh"
#include <ell/State.hh>
#include <biu/Matrix.hh>
#include <vector>

bool compareSP2(const ell::State* a, const ell::State* b);

namespace ell {
	 

class BarrierTree: private BarrierTree_IO, public LandscapeTopology {
protected:
	std::vector<const State*> vMinima;
	
public:
	BarrierTree(double delta = 0.0003877872621);
	virtual ~BarrierTree();
	
	///////////////// LANDSCAPE TOPOLOGY INTERFACE //////////////////
	//! sets a Saddle State between Minima with indices i and j
	virtual bool 
	addSaddle(const size_t i, const size_t j, const State* const s);
	
	//! sets a Saddle State between Minima States i and j
	virtual bool 
	addSaddle(const State* const i, const State* const j, const State* const s);
	
	//! adds a minimum State and returns the index of this minimum 
	virtual const size_t 
	addMin(const State* const i);
	
	//! returns index of State if it's a minimum, returns INT_MAX if State is no known minimum 
	virtual const size_t 
	getMinIndex(const State* const i) const;
	
	//! returns number of stored minima
	virtual const size_t 
	getMinCount() const;
	
	//! returns State of minimum with given index, returns NULL if index is out of bounds
	virtual const State* const 
	getMin(size_t i) const;
	
	//! returns the internal vector of minima as a const reference
	virtual const std::vector<const State*> 
	getAllMin() const;
	
	//! calculates the RMSD between this and a given BarrierTree(can have different # of minima)
	virtual double 
	getRMSD(const LandscapeTopology* const) const;

	///////////////// ADDITIONAL BARRIER TREE INTERFACE //////////////////	
	inline
	bool 
	isLeaf(const State* s) const;
	
	inline
	bool 
	isSameLeaf(const size_t i, const size_t j) ;
		
	//! returns energy of Saddle between minima mit indices i and j
	double
	getBarrier(const size_t, const size_t);

	void
	addBarrier(const State* s1, const State* s2, double energy);	

//	//! returns State* of Saddle between minima mit indices i and j
//	const State* 
//	getBarrierState(const size_t, const size_t) ;
	
	//! returns medium energy of all saddles
	const double
	getMediumBarrier() const;

	//! Calculates the sum of all barrier energies
	//! @return barrier energy sum
	const double
	getBarrierSum() const;

	void 
	printStatusInformation(std::ostream& out) const;
	
	void 
	printBarriers(std::ostream& out) const;
	
	void 
	sort();
	
	bool
	check();
	
	const size_t 
	getLeafCount() const;
};

}//ell


////! Reads the elements of the network(minima, saddle heights) from a file stream.
////! @param in the input stream to read from
////! @param sn the BarrierTree to fill
////! @return the read stream 
//std::ifstream&
//operator>>(std::ifstream& in,  ell::BarrierTree& sn);

//! Reads the elements of the network(minima, saddle heights) from a stream.
//! @param in the input stream to read from
//! @param sn the BarrierTree to fill
//! @return the read stream 
std::istream&
operator>>(std::istream& in,  ell::BarrierTree& sn);

////! Prints the saddlenet with structure und saddleenergies to stream
////! resulting file is a graphviz file, see http://www.graphviz.org
////! @param out the output stream to write to
////! @param sn the BarrierTree to print
////! @return the changed output stream out
//std::ofstream&
//operator << (std::ofstream& out, const ell::BarrierTree& sn);

//! Prints the saddlenet as pairs of Saddleindices and energy to outstream
//! @param out the output stream to write to
//! @param sn the BarrierTree to print
//! @return the changed output stream out
std::ostream&
operator << (std::ostream& out, ell::BarrierTree& sn);

#endif /*BARRIERTREE_HH_*/
