#include "ell/tp/BarrierTree_Advanced.hh"
#include <cassert>
#include <map>		// for calculateSampledRMSD(..) and sortSubtree(..)
#include <iomanip>


BarrierTree_Advanced::BarrierTree_Advanced(double _beta, int _mergingThreshold)
 : BarrierTree_Basic(_beta, _mergingThreshold) {
}

BarrierTree_Advanced::BarrierTree_Advanced(double _beta, int _mergingThreshold,
	const StateVec& localMinVec)
 : BarrierTree_Basic(_beta, _mergingThreshold, localMinVec) {
}

BarrierTree_Advanced::BarrierTree_Advanced(const BarrierTree_Advanced& bT2)
 : BarrierTree_Basic(bT2) {
}

BarrierTree_Advanced::~BarrierTree_Advanced() {
}

double
BarrierTree_Advanced::calculateSaddleHeightSum(Node* root) const {

	assert(root != NULL);
	
	double saddleEsum = 0.0;

	// root is an internal node
	if (!isLeaf(root)) {
		saddleEsum = ((double)root->energy / 100);
		  // init sum with first child
		saddleEsum += calculateSaddleHeightSum(root->leftChild);
		
		Node* sibling = root->leftChild->rightSibling;
		
		while (sibling != NULL) {
			saddleEsum += calculateSaddleHeightSum(sibling);
			sibling = sibling->rightSibling;
		}
	}
	
	return saddleEsum;
}

std::pair<double,size_t>
BarrierTree_Advanced::calculateSaddleHeightSumLeaves(Node* root) const {

	assert(root != NULL);
	
	double saddleEsum = 0.0;
	size_t leaves = 1;
	double multiplier = 1.0;

	// root is an internal node
	if (!isLeaf(root)) {
		  // init sum with first child
		Node* sibling = root->leftChild;
		
		std::pair<double,size_t> res = calculateSaddleHeightSumLeaves(sibling);
		saddleEsum	+= res.first;
		multiplier	= 0.0;
		leaves 		= res.second;;
		sibling = sibling->rightSibling;
		
		while (sibling != NULL) {
			std::pair<double,size_t> res = calculateSaddleHeightSumLeaves(sibling);
			saddleEsum	+= res.first;
			multiplier	+= leaves * res.second;
			leaves 		+= res.second;;

			sibling = sibling->rightSibling;
		}
		saddleEsum += (double)root->energy  * multiplier;
	}
	
	return std::pair<double,size_t>(saddleEsum,leaves);
}


double
BarrierTree_Advanced::calculateAverageSaddleHeight() const {
	size_t numOfLocalMins = getNumOfLocalMins();
	
	double divider = (double)(numOfLocalMins*(numOfLocalMins-1)) / 2.0;

	if ( numOfLocalMins < 2) { // no barrier available
		return (double)(INT_MAX);
	}
	  // get saddle energy sum
	std::pair<double,size_t> res = calculateSaddleHeightSumLeaves(barrierTreeRoot);

	  // calculate average value
	return (res.first / divider);
	
}


//double
//BarrierTree_Advanced::calculateAverageSaddleHeight() const {
//		// sum of saddle heights
//	double sumOfHeights = 0;
//	size_t numOfLocalMins = getNumOfLocalMins();
//
//		// sum up saddle heights
//	for (size_t i = 0; i < numOfLocalMins; i++) {
//		for (size_t j = i+1; j < numOfLocalMins; j++) {
//			sumOfHeights += getLCAInfo(getLeaf(index2StateInfo[i].state),
//						getLeaf(index2StateInfo[j].state)).lca->energy;
//		}
//	}
//	
//	std::cerr <<" sum = "  <<std::setw(15) <<std::fixed <<std::setprecision(2) <<sumOfHeights <<std::endl;
//
//	sumOfHeights /= static_cast<double>(
//					 numOfLocalMins * (numOfLocalMins - 1) / 2 );
//	return sumOfHeights;
//
//}

double
BarrierTree_Advanced::calculateRMSD(const BarrierTree_Advanced& bT2,
	int maxBarrierHeight) const {
		// the root mean square deviation
	double rmsd = 0.;
		// number of local mins in this barrier tree
	size_t numOfLocalMins = getNumOfLocalMins();
		// mapping from the local min indices of this to the analog local min
		// indices of barrier tree bT2
		// element i is BarrierTree_Basic::INVALID_INDEX if local min with
		// index i of this has no analog local min in barrier tree bT2
	std::vector<size_t> analogLocalMinIndexInBT2(numOfLocalMins,
							BarrierTree_Basic::INVALID_INDEX);
		// boltzmann weights of the local mins of this barrier tree
	std::vector<double> boltzmannWeightsOfLocalMinsOfThisBT;
		// set element i false if local min with index i of bT2 has analog
		// local min in this barrier tree
	std::vector<bool> noAnalogLocalMinOfBT2(bT2.getNumOfLocalMins(), true);
		// boltzmann weights of the local mins of bT2
	std::vector<double> boltzmannWeightsOfLocalMinsOfBT2;
		// number of local mins of bT2 having no analog local min in this
		// barrier tree
	size_t numOfNoAnalogLocalMinOfBT2 = noAnalogLocalMinOfBT2.size();
		// saddle heights in this and barrier tree bT2
	int height1, height2;
		// weight for saddle heights (= max of Boltzmann weights of
		// both connected states)
	double bWeight;
	size_t i, j, i2, j2;

		// if maxBarrierHeight hasn't been passed, set it to the maximal
		// saddle height (= energy of root) of both barrier trees
	if (maxBarrierHeight == BarrierTree_Basic::INVALID_ENERGY) {
		barrierTreeRoot->energy >= bT2.barrierTreeRoot->energy ?
		 maxBarrierHeight = barrierTreeRoot->energy : 
		 maxBarrierHeight = bT2.barrierTreeRoot->energy;
	}

		// fill analogLocalMinIndexInBT2, noAnalogLocalMinOfBT2 and
		// boltzmannWeightsOfLocalMinsOfThisBT
	for (i = 0; i < numOfLocalMins; i++) {
		if (bT2.isInTree(index2StateInfo[i].state)) {
			analogLocalMinIndexInBT2[i] =
			 bT2.getLocalMinIndex(index2StateInfo[i].state);
			noAnalogLocalMinOfBT2[analogLocalMinIndexInBT2[i]] = false;
			numOfNoAnalogLocalMinOfBT2--;
		}
		boltzmannWeightsOfLocalMinsOfThisBT.push_back(
			calculateBoltzmannWeight(i));
	}
		// fill boltzmannWeightsOfLocalMinsOfBT2
	for (i = 0; i < bT2.getNumOfLocalMins(); i++) {
		boltzmannWeightsOfLocalMinsOfBT2.push_back(
			bT2.calculateBoltzmannWeight(i));
	}

		// calculate rmsd
	for (i = 0; i < numOfLocalMins; i++) {
		for (j = i+1; j < numOfLocalMins; j++) {
			height1 = getLCAInfo(getLeaf(index2StateInfo[i].state),
						getLeaf(index2StateInfo[j].state)).lca->energy;
			i2 = analogLocalMinIndexInBT2[i];
			j2 = analogLocalMinIndexInBT2[j];

			assert(i < boltzmannWeightsOfLocalMinsOfThisBT.size()
					&& j < boltzmannWeightsOfLocalMinsOfThisBT.size());

			bWeight = std::max(boltzmannWeightsOfLocalMinsOfThisBT[i],
								boltzmannWeightsOfLocalMinsOfThisBT[j]);
			if ( i2 != BarrierTree_Basic::INVALID_INDEX
			      && j2 != BarrierTree_Basic::INVALID_INDEX) {
				height2 = bT2.getLCAInfo(
							bT2.getLeaf(bT2.index2StateInfo[i2].state),
							bT2.getLeaf(bT2.index2StateInfo[j2].state)).
							lca->energy;
				rmsd += (bWeight * pow( (height1 - height2), 2));
			}
			else {
				rmsd += (bWeight * pow( (height1 - maxBarrierHeight), 2));
			}
		}
	}

	for (i = 0; i < noAnalogLocalMinOfBT2.size(); i++) {
		for (j = i+1; j < noAnalogLocalMinOfBT2.size(); j++) {
				// i or j have to be true
				// (= no analog local min of a local min of this barrier tree)
			if ( noAnalogLocalMinOfBT2[i] || noAnalogLocalMinOfBT2[j] ) {

				assert(i < boltzmannWeightsOfLocalMinsOfBT2.size()
						&& j < boltzmannWeightsOfLocalMinsOfBT2.size());

				bWeight = std::max(boltzmannWeightsOfLocalMinsOfBT2[i],
									boltzmannWeightsOfLocalMinsOfBT2[j]);
				height2 = bT2.getLCAInfo(
							bT2.getLeaf(bT2.index2StateInfo[i].state),
							bT2.getLeaf(bT2.index2StateInfo[j].state)).
							lca->energy;
				rmsd += (bWeight * pow( (maxBarrierHeight - height2), 2));
			}
		}
	}

	if ( numOfLocalMins > 1 || numOfNoAnalogLocalMinOfBT2 > 1 ) {
		rmsd /= static_cast<double>(
				 (numOfLocalMins + numOfNoAnalogLocalMinOfBT2) *
 				 (numOfLocalMins + numOfNoAnalogLocalMinOfBT2 - 1) / 2 );
	}

	return sqrt(rmsd);
}

double
BarrierTree_Advanced::calculateSampledRMSD(const BarrierTree_Advanced& bT2,
	size_t numOfSampledLocMins, int maxBarrierHeight) const {
		// choose the randomly chosen Boltzmann distributed local minimum
		// states from the larger barrier tree
	if (this->getNumOfLocalMins() < bT2.getNumOfLocalMins()) {
		return bT2.getSampledRMSD(*this, numOfSampledLocMins, maxBarrierHeight);
	}
	else {
		return this->getSampledRMSD(bT2, numOfSampledLocMins, maxBarrierHeight);
	}
}

double
BarrierTree_Advanced::getSampledRMSD(const BarrierTree_Advanced& bT2,
	size_t numOfSampledLocMins, int maxBarrierHeight) const {
		// numOfSampledLocMins has to be more than the number of local minimum
		// states, otherwise the result of calculateRMSD will be returned
	if (numOfSampledLocMins >= getNumOfLocalMins()) {
		return calculateRMSD(bT2, maxBarrierHeight);
	}

		// the sampled root mean square deviation
	double sampledRmsd = 0.;
		// indices of the sampled local mins of this barrier tree, their
		// boltzmann weights and the analog local min index of barrier tree bT2
		// if it has an analog local min in bT2, else analog local min index
		// is BarrierTree_Basic::INVALID_INDEX
	std::map< size_t, std::pair<double, size_t> > sampledLocalMinsOfThisBT;
		// saddle heights in this and barrier tree bT2
	int height1, height2;
		// weight for saddle heights (= max of Boltzmann weights of
		// both connected states)
	double bWeight;
	std::map<size_t, std::pair<double, size_t> >::const_iterator it1, it2;
	size_t i, i2;

		// if maxBarrierHeight hasn't been passed, set it to the maximal
		// saddle height (= energy of root) of both barrier trees
	if (maxBarrierHeight == INVALID_ENERGY) {
		barrierTreeRoot->energy >= bT2.barrierTreeRoot->energy ?
		 maxBarrierHeight = barrierTreeRoot->energy : 
		 maxBarrierHeight = bT2.barrierTreeRoot->energy;
	}

		// fill sampledLocalMinsOfThisBT
	while (sampledLocalMinsOfThisBT.size() < numOfSampledLocMins) {
		i = getLocalMinIndex(getBoltzmannDistributedRandomLocalMin());
		if (sampledLocalMinsOfThisBT.find(i) == sampledLocalMinsOfThisBT.end()) {
			bT2.isInTree(index2StateInfo[i].state) ?
			 i2 = bT2.getLocalMinIndex(index2StateInfo[i].state) :
			 i2 = BarrierTree_Basic::INVALID_INDEX;
			sampledLocalMinsOfThisBT.insert(std::make_pair(
				i, std::make_pair(calculateBoltzmannWeight(i), i2)));
		}
	}

		// calculate sampledRmsd
	for (it1 = sampledLocalMinsOfThisBT.begin();
		   it1 != sampledLocalMinsOfThisBT.end(); it1++) {
		for (it2 = it1, it1++; it2 != sampledLocalMinsOfThisBT.end(); it2++) {
			height1 = getLCAInfo(getLeaf(index2StateInfo[it1->first].state),
						getLeaf(index2StateInfo[it2->first].state)).
						 lca->energy;
			bWeight = std::max(it1->second.first, it2->second.first);
			if ( it1->second.second != BarrierTree_Basic::INVALID_INDEX
			      && it2->second.second != BarrierTree_Basic::INVALID_INDEX) {
				height2 = bT2.getLCAInfo(
							bT2.getLeaf(bT2.index2StateInfo[
								it1->second.second].state),
							bT2.getLeaf(bT2.index2StateInfo[
								it2->second.second].state)).lca->energy;
				sampledRmsd += (bWeight * pow( (height1 - height2), 2));
			}
			else {
				sampledRmsd += (bWeight *
									pow( (height1 - maxBarrierHeight), 2));
			}
		}
	}

	if ( numOfSampledLocMins > 1 ) {
		sampledRmsd /= static_cast<double>(
					numOfSampledLocMins * (numOfSampledLocMins - 1) / 2 );
	}

	return sqrt(sampledRmsd);
}

void
BarrierTree_Advanced::reduceBarrierTree(size_t reducedToLeaves) {
// hier koennte man eine clevere nicht-"brut-force"-Implementierung verwenden
		// leaf which will be deleted
	Node *deletedLeaf, *parent;
		// save temporarily pointer of deletedStates to free the memory
	StateVec deletedStates;

	while (getNumOfLeaves() > reducedToLeaves) {
		deletedLeaf = leafIndex2Leaf[getNumOfLeaves()-1];
		parent = deletedLeaf->parent;
		removeChildFromParent(deletedLeaf);
		consolidateAfterLeafRemoval(parent);
		for(StateVec::const_iterator cit =
			 leafIndex2StateVec[deletedLeaf->leafIndex].begin();
			 cit !=  leafIndex2StateVec[deletedLeaf->leafIndex].end();
			 cit++) {
			deletedStates.push_back(*cit);
		}
		deleteLeaf(deletedLeaf);
		for(StateVec::const_iterator cit =	 deletedStates.begin();
			 cit !=  deletedStates.end(); cit++) {
			
			delete *cit;
		}
		deletedStates.clear();
	}
}

void
BarrierTree_Advanced::sortBarrierTree() {
	if (barrierTreeRoot) {
		sortLocalMinsByEnergy();
		sortSubtree(barrierTreeRoot);
	}
}

void
BarrierTree_Advanced::sortLocalMinsByEnergy() {
	size_t numOfLocalMins = getNumOfLocalMins();
	size_t numOfLeaves = getNumOfLeaves();
		// old and new index of the current leaf
	size_t oldLeafIndex, newLeafIndex = 0;
		// mapping from old leaf index to flag indicating if the leaf has
		// already been updated
	std::vector<bool> oldLeafUpdated(numOfLeaves, false);
		// updated leafIndex2Leaf with updated leaf indices
	NodeVec	sortedLI2L(numOfLeaves, NULL);
	StateIndexMap::iterator it;

		// sort the local minimum states by their energy
	stable_sort(index2StateInfo.begin(), index2StateInfo.end(),
		    StateInfoComp());

		// delete all entries of leafIndex2StateVec
	leafIndex2StateVec.clear();
	leafIndex2StateVec.resize(numOfLeaves, StateVec());

		// update local min indices in StateIndexMap and the leaf indices
		// according to the new local minima order
	for (size_t i = 0; i < numOfLocalMins; i++) {
			// find state in StateIndexMap
		it = State2Index.find(index2StateInfo[i].state);
			// update index in StateIndexMap
		it->second = i;

		oldLeafIndex = index2StateInfo[i].leafIndex;
			// update leaf index
				// leaf hasn't been updated yet
		if (!(oldLeafUpdated[oldLeafIndex])) {

			assert(newLeafIndex < numOfLeaves);

				// add new leaf index and the corresponding leaf to sortedLI2L
			sortedLI2L[newLeafIndex] = leafIndex2Leaf[oldLeafIndex];
				// update leaf index at the corresponding leaf
			sortedLI2L[newLeafIndex]->leafIndex = newLeafIndex;
			oldLeafUpdated[oldLeafIndex] = true;
				// update index2StateInfo with new leaf index
			index2StateInfo[i].leafIndex = newLeafIndex;
				// add local min in leafIndex2StateVec for new leaf index
			leafIndex2StateVec[newLeafIndex].push_back(
				index2StateInfo[i].state);
			newLeafIndex++;
		}
				// leaf has already been updated
		else {
				// update index2StateInfo with leaf index from leaf being
				// already updated
			index2StateInfo[i].leafIndex =
				leafIndex2Leaf[oldLeafIndex]->leafIndex;
				// add local min in leafIndex2StateVec for the updated
				// leaf index
			leafIndex2StateVec[index2StateInfo[i].leafIndex].push_back(
				index2StateInfo[i].state);
		}
	}

	leafIndex2Leaf = sortedLI2L;

}

size_t
BarrierTree_Advanced::sortSubtree(Node* root) {

	assert(root != NULL);

		// return leaf index if root is a leaf
	if (isLeaf(root)) {
			return root->leafIndex;
	}
		// recursive call of sortSubtree if root is an internal node
	else {
		Node* sibling = root->leftChild->rightSibling;
			// node has 2 children -> sort children by direct comparison
		if (sibling->rightSibling == NULL)	{
			size_t leftChildIndex = sortSubtree(root->leftChild);
			size_t siblingIndex = sortSubtree(sibling);
				// left child has to have higher leaf index
			if (leftChildIndex < siblingIndex) {
					// exchange positions of both children
				Node* oldLeftChild = root->leftChild;
				root->leftChild = sibling;
				sibling->leftSibling = NULL;
				sibling->rightSibling = oldLeftChild;
				oldLeftChild->leftSibling = sibling;
				oldLeftChild->rightSibling = NULL;
				siblingIndex = leftChildIndex;
			}
				// return siblingIndex which is minimal leaf index of subtree
			return siblingIndex;
		}	
			// node has more than 2 children
			// -> sort children using a sorted map
		else {
			std::map<size_t, Node*, std::less<size_t> > sortedChildren;
				// insert all children in a map
			sortedChildren.insert(
			 std::make_pair(sortSubtree(root->leftChild), root->leftChild));
			while (sibling != NULL) {
				sortedChildren.insert(
				 std::make_pair(sortSubtree(sibling), sibling));
				sibling = sibling->rightSibling;
			}
				// child with highest leaf index is the last element
			std::map<size_t, Node*>::reverse_iterator rit =
				sortedChildren.rbegin();
				// child with highest leaf index is the left child
			root->leftChild = rit->second;
			rit->second->leftSibling = NULL;
			rit++;
				// children are sorted from the left to the right
				// by their leaf indices in decreasing order
			while (rit != sortedChildren.rend()) {
				rit->second->leftSibling = (--rit)->second;
				rit->second->rightSibling = (++rit)->second;
				rit++;
			}
			(--rit)->second->rightSibling = NULL;
				// return leaf index of rightmost child
			return rit->first;
		}
	}
}
