#ifndef BARRIERTREE_ADVANCED_HH_
#define BARRIERTREE_ADVANCED_HH_


#include "ell/tp/BarrierTree_Basic.hh"


		/*! This class extends the barrier tree class with basic operations by
		 *  advanced functions.
		 *
		 *  @author Andreas Richter
		 */


class BarrierTree_Advanced : public BarrierTree_Basic
{
protected:
///////////////////////////////////////////////////////////////////////////////
// Function objects (functors)
		/*! Compares the energy values of two entries of index2StateInfo.
		 *  Compares the cStates in the StateInfo type lexicographically
		 *  if they have the same energy value.
		 */ 
	struct	StateInfoComp {
		bool operator()(const StateInfo& a, const StateInfo& b) const {
			return (a.energy != b.energy) ? (a.energy < b.energy) :
					 (a.state->toString() < 
					   b.state->toString());
		}
	};
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Member functions
		/*! Calculates the root mean square deviation of the current and the
		 *  passed barrier tree weighted with Boltzmann factors. To speed up
		 *  the calculation, only numOfSampledLocMins many local minimum
		 *  states are randomly chosen Boltzmann distributed from the current
		 *  barrier tree.
		 *  The maximal possible barrier height will be used if a barrier
		 *  of one has no analog in the other tree. If it isn't passed,
		 *  the maximal barrier height out of both trees will be taken.
		 *  @param numOfSampledLocMins the number of sampled local minimum
		 *   states has to be more than the number of local minimum states,
		 *	 otherwise the result of calculateRMSD will be returned
		 */
	double		getSampledRMSD(const BarrierTree_Advanced& bT2,
					size_t numOfSampledLocMins,
					int maxBarrierHeight = INVALID_ENERGY) const;

		/*! Sorts the indices of the local minimum states by their energy
		 *  values in increasing order. The indices of the leaves will be
		 *  updated in the way that the lowest energy state is
		 *  referenced by the leaf with the lowest leaf index. The list of
		 *  cStates belonging to a leaf will be sorted as well by increasing
		 *  energy values.
		 */
	void		sortLocalMinsByEnergy();

		/*! Sorts the subtree below the passed node root. Each node has
		 *  always a lower leaf index than its left sibling. Returns
		 *  the lowest leaf index within the subtree.
		 */
	size_t		sortSubtree(Node* root);
///////////////////////////////////////////////////////////////////////////////

		/*! Calculates recursively the saddle height sum of a subtree of the 
		 * barrier tree.
		 * @param root root of the subtree
		 * @return the sum the saddle energies
		 */
	double		calculateSaddleHeightSum(Node* root) const;
	std::pair<double,size_t>
	calculateSaddleHeightSumLeaves(Node* root) const;

public:
	BarrierTree_Advanced(double _beta, int _mergingThreshold);

	BarrierTree_Advanced(double _beta, int _mergingThreshold,
		const StateVec& localMinVec);

	BarrierTree_Advanced(const BarrierTree_Advanced& bT2);

	virtual ~BarrierTree_Advanced();


///////////////////////////////////////////////////////////////////////////////
// Member functions
		/*! Calculates the average saddle height (energy) of the barrier tree.
		 * @return The average saddle energy or INT_MAX if less than 2 minima 
		 *         are present.
		 */
	double		calculateAverageSaddleHeight() const;

		/*! Calculates the root mean square deviation of the current
		 *  and the passed barrier tree weighted with Boltzmann factors.
		 *  The maximal possible barrier height will be used if a barrier
		 *  of one has no analog in the other tree. If it isn't passed,
		 *  the maximal barrier height out of both trees will be taken. 
		 */
	double		calculateRMSD(const BarrierTree_Advanced& bT2,
					int maxBarrierHeight = BarrierTree_Basic::INVALID_ENERGY)
					const;

		/*! Calculates the root mean square deviation of the current and the
		 *  passed barrier tree weighted with Boltzmann factors. To speed up
		 *  the calculation, only numOfSampledLocMins many local minimum
		 *  states are randomly chosen Boltzmann distributed from the current
		 *  barrier tree or bT2, depending on which barrier tree is larger.
		 *  The maximal possible barrier height will be used if a barrier
		 *  of one has no analog in the other tree. If it isn't passed,
		 *  the maximal barrier height out of both trees will be taken.
		 *  @param numOfSampledLocMins the number of sampled local minimum
		 *   states has to be more than the number of local minimum states,
		 *	 otherwise the result of calculateRMSD will be returned
		 */
	double		calculateSampledRMSD(const BarrierTree_Advanced& bT2,
					size_t numOfSampledLocMins,
					int maxBarrierHeight = BarrierTree_Basic::INVALID_ENERGY)
					const;

		/*! Reduces the barrier tree to reducedToLeaves many leaves preserving
		 *  the topology of the  barrier tree. The leaves with leaf indices in
		 *  [0, reducedToLeaves) will be kept.
		 *  Note that the barrier tree has to be sorted to get the lowest
		 *  energy leaves.
		 */
	void		reduceBarrierTree(size_t reducedToLeaves);

		/*! Sorts the local minimum states by their energy.
		 *  Sorts the leaves of the barrier tree by the leaf indices.
		 *  Each node has always a lower leaf index than its left
		 *  sibling. The lowest energy state is referenced by the leaf
		 *  with the lowest leaf index.
		 *  The tree is sorted breadth-first starting with the 
		 *  root of the barrier tree.
		 */
	void		sortBarrierTree();
//////////////////////////////////////////////////////////////////////////////
};

#endif /*BARRIERTREE_ADVANCED_HH_*/
