#ifndef BARRIERTREE_BASIC_HH_
#define BARRIERTREE_BASIC_HH_
// switch off debugging
//#define NDEBUG


#include <ell/State.hh>
#include <vector>
#include <ext/hash_map>
#include <ext/hash_set>
#include <cassert>
#include <cmath> // for calculateBoltzmannWeight(..)
#include <iostream>
#include "ell/util/DJBHash.hh"
#include "ell/StatePtrEq.hh"

		/*! This class saves the local minimum states in compressed
		 *  representation. It also provides the barrier tree with leaves
		 *  and their associated local minimum states and the saddle
		 *  heights as internal nodes of the tree. This class allows
		 *  basic operations on this barrier tree.
		 *
		 *  @author Andreas Richter
		 */


typedef std::vector<const ell::State*>		StateVec;

class BarrierTree_Basic
{
protected:
///////////////////////////////////////////////////////////////////////////////
// Nested classes
		/*! This class provides a node labelled with an energy value.
		 *  If the node is a leaf, it has a leaf index. Otherwise the leaf
		 *  index will have the value INVALID_INDEX.
		 */
	class Node
	{
	public:
			//! Pointer to parent, left child and siblings of the node
		Node	*parent, *leftChild, *rightSibling, *leftSibling;
			//! Energy saved in the node
		int	energy;
			//! Index of the leaf
		size_t leafIndex;
			//! Constructor
		Node(int _energy = INVALID_ENERGY, size_t _leafIndex = INVALID_INDEX)
		: parent(NULL), leftChild(NULL), rightSibling(NULL), leftSibling(NULL),
		  energy(_energy), leafIndex(_leafIndex) {}
	};

		/*! This class contains information about the least common ancestor
		 *  of two nodes like a pointer to the lca and the distances from
		 *  each of the nodes to the lca.
		 */
	class LCAInfoType
	{
	public:
			//! The least common ancestor of both nodes.
		Node*	lca;
			//! Distance from the node to the lca.
		size_t	distNode1ToLCA, distNode2ToLCA;
			//! Constructor
		LCAInfoType(Node* _lca, size_t _distNode1ToLCA, size_t _distNode2ToLCA)
		: lca(_lca), distNode1ToLCA(_distNode1ToLCA),
		  distNode2ToLCA(_distNode2ToLCA) {}
	};

		/*! This class contains a pointer to a compressed local minimum state,
		 *  the energy of this state and the index of the leaf it belongs to.
		 */
	class StateInfo
	{
	public:
		const ell::State*	state;
		int			energy;
		size_t			leafIndex;
			//! Constructor
		StateInfo(const ell::State* _State = NULL,
		 int _energy = INVALID_ENERGY, size_t _leafIndex = INVALID_INDEX)
		 : state(_State), energy(_energy), leafIndex(_leafIndex) {}
	};
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Type defs
	typedef std::vector<Node*> 				NodeVec;
	typedef std::vector<const ell::State*>	StateVec;
	typedef std::vector<StateInfo>			StateInfoVec;
	typedef std::vector<StateVec>				IndexStateVec;
	typedef __gnu_cxx::hash_map<const ell::State*, size_t, DJBHash,
			StatePtrEq>						StateIndexMap;
	typedef __gnu_cxx::hash_set<const ell::State*, DJBHash, StatePtrEq>
												StateSet;
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Constants
		//! Constant that represents an invalid energy value.
	static const int INVALID_ENERGY;
		//! Constant that represents an invalid index value.
	static const size_t INVALID_INDEX;
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Data members
		/*! a mapping from an index of a local minimum state to the compressed
		 *  local minimum state, its energy and the corresponding leaf index
		 */
	StateInfoVec	index2StateInfo;

		/*! a mapping from a compressed local minimum state to the corresponding
		 *  index
		 */
	StateIndexMap	State2Index;

		/*! a mapping from a leaf index to the leaf with this index
		 */
	NodeVec			leafIndex2Leaf;

		/*! a mapping from a leaf index to a vector of states belonging to
		 *  this leaf
		 */
	IndexStateVec	leafIndex2StateVec;
public:
		/*! compressed local minimum states which were removed from the 
		 *  barrier tree since they turned out to be shoulder points
		 */
	StateSet		shoulderPointStates;
protected:
		/*! a pointer to the root of the barrier tree
		 */
	Node*			barrierTreeRoot;

		/*! The thermodynamic beta ( = 1/(k*T) ) of the Boltzmann distribution.
		 */
	double			beta;

		/*! Represents the maximal energy barrier between two leaves
		 *  having the same energy to merge the two leaves.
		 */ 
	int				mergingThreshold;
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Member functions
		/*! Returns the Boltzmann factor of the local minimum state with the
		 *  passed index.
		 */
	double		calculateBoltzmannWeight(size_t index) const;

		/*! Consolidates the barrier tree after the removal of a leaf.
		 *  More precisely, internal nodes having just one child will
		 *  be removed and the child will be appended to their parent.
		 *  @param parent parent of the leaf which has been removed
		 */
	void		consolidateAfterLeafRemoval(Node* parent);

		/*! Deletes deletedLeaf which has to be a leaf from the tree.
		 *  Removes its data from the local minimum states and updates
		 *  all data structures.
		 */
	void		deleteLeaf(Node* deletedLeaf);

		/*! Returns the index of the passed local minimum state if it is saved
		 *  in the tree, else returns INVALID_INDEX.
		 */
	size_t		getLocalMinIndex(const ell::State* state) const;

		/*! Returns information about the least common ancestor (LCA)
		 *  of node1 and node2.
		 */
	LCAInfoType	getLCAInfo(Node* node1, Node* node2) const;

		/*! Returns a pointer to the leaf which references the passed compressed
		 *  local minimum state if it is saved in the tree, else returns NULL.
		 */
	Node*		getLeaf(const ell::State* state) const;
public:
		/*! Returns the number of leaves of the barrier tree.
		 */
	size_t		getNumOfLeaves() const;

		/*! Returns the number of saved local minimum states.
		 */
	size_t		getNumOfLocalMins() const;

protected:
		/*! Returns true if the passed node is a leaf of the barrier tree,
		 *  else returns false.
		 */
	bool		isLeaf(const Node* node) const;

		/*! Removes child node from its parent node.
		 */
	void		removeChildFromParent(Node* child) const;
///////////////////////////////////////////////////////////////////////////////


private:
///////////////////////////////////////////////////////////////////////////////
// Member functions
		/*! Adds the saddle height between an internal node, which
		 *  has to be the highest node below the saddleHeight, and
		 *  a subtree, which isn't allowed to be connected to the subtree
		 *  containing the internal node.
		 */
	void		addSaddleHeightBetweenInternalNodeAndSubtree(
					Node* internalNode, Node* subtreeRoot, int saddleHeight);

		/*! Appends child node to parent node and makes it the left child if it
		 *  doesn't exist or else the right sibling of the left child.
		 */
	void		appendChildToParent(Node* parent, Node* child) const;

		/*! Returns true if the two passed leaves have the same energy and if
		 *  the energy barrier between them is below or equal to 
		 *  mergingThreshold.
		 */
	bool		areSameEnergyLeavesToBeMerged(const Node* leaf1,
					const Node* leaf2, int energyBarrier) const;

		/*! Clones the the passed root node of the origin barrier tree and the
		 *  subtrees below it and fills the passed leafIndex2LeafClone of the
		 *  clone recursively.
		 */
	Node*		cloneNode(Node* bT2Root, NodeVec& leafIndex2LeafClone);

		/*! Updates the leaf indices of all local minimum states belonging
		 *  to deletedLeaf to the leaf index of mergedLeaf and deletes
		 *  the leaf deletedLeaf.
		 */
	void		deleteAndMergeLeaf(Node* deletedLeaf, Node* mergedLeaf);

		/*! Deletes shoulderPoint which has to be a leaf from the tree.
		 *  Inserts shoulderPoint's data into shoulderPointStates and
		 *  calls deleteLeaf(shoulderPoint).
		 */
	void		deleteShoulderPointNode(Node* shoulderPoint);

		/*! Frees the memory from the passed root node and the subtrees below
		 *  it recursively.
		 */
	void		freeNode(Node* root);

		/*! Frees the memory from all local minimum States and all shoulder
		 *  point States.
		 */
	void		freeMinDataStructures();

		/*! Returns a pointer to the highest ancestor of the passed node
		 *  with highest ancestor's energy less than the
		 *  passed energyBound.
		 *  Returns the passed node if such an ancestor doesn't exist.
		 */
	Node*		getHighestAncestorBelowEnergyBound(Node* node, int energyBound)
					const;

		/*! Returns true if the passed node connected to the passed saddle
		 *  height is a shoulder point (same energy as the saddle height).
		 */
	bool		isShoulderPoint(const Node* node, int saddleHeight) const;

		/*! Returns true, if the passed leaf has been merged with a child of
			the passed parent, since areSameEnergyLeavesToBeMerged(..)
			returned true. Else return false.
		 */
	bool		mergeLeafWithChildOfParent(Node* parent, Node* leaf);

		/*! Replaces oldNode by newNode and removes oldNode from its parent
		 *  (if there is one).
		 */
	void		replaceNodeByNewNode(Node* oldNode, Node* newNode) const;
///////////////////////////////////////////////////////////////////////////////


public:
	BarrierTree_Basic(double _beta, int _mergingThreshold);

	BarrierTree_Basic(double _beta, int _mergingThreshold,
		const StateVec& localMinVec);

	BarrierTree_Basic(const BarrierTree_Basic& bT2);

	virtual ~BarrierTree_Basic();

	BarrierTree_Basic& operator=(const BarrierTree_Basic& bT2);
	
	enum UpdateEvent { UPDATED, MERGED2M2, M1_SHOULDER, M2_SHOULDER, NOTHING}; 


///////////////////////////////////////////////////////////////////////////////
// Member functions
		/*! Adds a new saddle height between the leaves referencing the
		 *  compressed states State1 and State2. One of this leaves has to be
		 *  a single leaf without a connection to any other nodes.
		 */
	void	addSaddleHeightToNewLeaf(const ell::State* state1,
				const ell::State* state2, int saddleHeight);

		/*! Returns a pointer to a randomly chosen Boltzmann distributed
		 *  compressed local minimum state of the barrier tree.
		 *  States which are more probable will be chosen more often.
		 */
	const ell::State*	getBoltzmannDistributedRandomLocalMin() const;

		/*! Returns a pointer to a randomly chosen compressed local minimum
		 *  state of the barrier tree.
		 */
	const ell::State*	getRandomLocalMin() const;

		/*! Inserts the localMinState whose energy is the passed energy and
		 *  a new leaf referencing this local minimum state.
		 *  Returns the new leaf.
		 */
	Node*	insertNewLocalMin(const ell::State* localMinState, int energy);

		/*! Returns true if the compressed local minimum state is in the
		 *  barrier tree, else return false.
		 */
	bool	isInTree(const ell::State* state) const;

		/*! Returns true if the compressed local minimum state is in the
		 *  barrier tree or if it was was removed since it's a shoulder
		 *  point, else return false.
		 */
	bool	isKnownLocalMin(const ell::State* state) const;

		/*! Updates the saddle height between the leaves referencing the
		 *  compressed states State1 and State2 if the new saddle height is
		 *  lower than the existing one.
		 * 
		 * @return	UPDATED if saddle height was just updated
		 * 			MERGED2M2 if min1 and min2 were merged due to too small barrier
		 * 			M1_SHOULDER min1 was removed from minima because it is a shoulder (no barrier to min2)
		 * 			M2_SHOULDER min2 was removed from minima because it is a shoulder (no barrier to min1)
		 * 			NOTHING no barrier update was done
		 */
	UpdateEvent
	updateSaddleHeight(const ell::State* min1,
				const ell::State* min2, int saddleHeight);
///////////////////////////////////////////////////////////////////////////////
};


///////////////////////////////////////////////////////////////////////////////
// Inline implementation
inline bool BarrierTree_Basic::areSameEnergyLeavesToBeMerged(const Node* leaf1,
				const Node* leaf2, int energyBarrier) const {
		// leaf1 and leaf2 have the same energy and the energy
		// barrier between them is below or equal to mergingThreshold
	return ( (leaf1->energy == leaf2->energy) &&
			  (std::abs(leaf1->energy - energyBarrier) <= mergingThreshold) );
}

inline double BarrierTree_Basic::calculateBoltzmannWeight(size_t index) const {
		// return Boltzmann factor of local minimum state with passed index
	return exp( -beta * static_cast<double>(index2StateInfo[index].energy) );
}

inline size_t BarrierTree_Basic::getLocalMinIndex(const ell::State* state)
				const {
		// search for the State in StateIndexMap
	StateIndexMap::const_iterator cit = State2Index.find(state);
		// return the index of the state if it's saved in the tree,
		// else return INVALID_INDEX
	return (cit != State2Index.end()) ? cit->second : INVALID_INDEX;
}

inline BarrierTree_Basic::Node* BarrierTree_Basic::getLeaf(
		const ell::State* state) const {
		// get index of the local min state
	size_t StateIndex = getLocalMinIndex(state);
		// return pointer to the leaf referencing the state
		// if the state is saved in the tree, else return NULL
	return (StateIndex != INVALID_INDEX) ?
			leafIndex2Leaf[index2StateInfo[StateIndex].leafIndex] : NULL;
}

inline size_t BarrierTree_Basic::getNumOfLeaves() const {
		return leafIndex2Leaf.size();
}

inline size_t BarrierTree_Basic::getNumOfLocalMins() const {
		return index2StateInfo.size();
}

inline const ell::State* BarrierTree_Basic::getRandomLocalMin() const {
		// return random local minimum state out of all local minimum states
		//	all local minimum state are uniformly distributed
	return index2StateInfo[(random() % getNumOfLocalMins())].state;
}

inline bool BarrierTree_Basic::isInTree(const ell::State* state) const {
		// return true if State is in the barrier tree
	return (getLocalMinIndex(state) != INVALID_INDEX);
}

inline bool BarrierTree_Basic::isKnownLocalMin(const ell::State* state)
				const {
		// returns true if State is in the barrier tree or if the State
		// was removed from the tree since it turned out to be a
		// shoulder point, else return false
	return (isInTree(state) ||
			  shoulderPointStates.find(state) != shoulderPointStates.end());
}

inline bool BarrierTree_Basic::isLeaf(const Node* node) const {
		// node is a leaf if its leaf index isn't INVALID_INDEX
	return (node->leafIndex != INVALID_INDEX);
}

inline bool BarrierTree_Basic::isShoulderPoint(const Node* node,
				int saddleHeight) const {
		// node is a shoulder point if it has the same energy like the saddleHeight
//	return node->energy == saddleHeight;
	//@TODO ONLY QUICKFIX !!! HAS TO BE UNDONE !!!
	return false;
}
///////////////////////////////////////////////////////////////////////////////

#endif /*BARRIERTREE_BASIC_HH_*/
