#include <biu/assertbiu.hh>
#include <ell/Walk.hh>
#include <ell/SC_Counting.hh>
#include <cmath>
#include <fstream>
#include <iostream>
#include <sstream>
#include <iostream>
#include "ell/tp/SaddleNet.hh"
#include "ell/fromString.hh"
using namespace std;
using namespace ell;

namespace ell {

size_t	SaddleNet::rejSaddles = 0;
size_t	SaddleNet::addSaddles = 0;

SaddleNet::~SaddleNet() {
	for (size_t i = 0; i < mAdjacency.numRows(); ++i) {
		for (size_t j = i; j < mAdjacency.numRows(); ++j) {
			delete mAdjacency[i][j];
			mAdjacency[i][j] = mAdjacency[j][i] = NULL;
		}
	}	
};


bool
SaddleNet::addSaddle(const size_t i, const size_t j, const State* const s) {
	bool changed = false;

	assertbiu(i < mLabel.numRows(), "index i out of bound!");
	assertbiu(j < mLabel.numRows(), "index j out of bound!");
	assertbiu(i != j, "i and j equal.. no saddle!");
	assertbiu(s != NULL, "State s is NULL pointer");
	
	if(s==NULL) {
		cout << "SaddleNet::addSaddle("<<i<<","<<j<<","<<s<<") ...."<<endl;
		return false;
	}
	
	//! indices valid?
	if(i < mLabel.numRows() && j < mLabel.numRows()) {
		//! check if saddle doesn't exist
		//! new found saddle energy higher than given by network?
		if((mAdjacency[i][j] == NULL) ||
				mAdjacency[i][j]->getEnergy() > s->getEnergy()) {
			++addSaddles;
			delete mAdjacency[i][j];
			mAdjacency[i][j] 	= mAdjacency[j][i] = s->clone();
			mLabel[i][j] 		= mLabel[j][i] 	= s->getEnergy();
			changed = true;
		} else rejSaddles++;
	} else cout << "minimum "<<max(i,j)<<" doesn't exist!"<<endl;
	
	return changed;
}

bool
SaddleNet::addSaddle(const State* const si, const State* const sj, const State* const s) {
//	std::cerr <<"SaddleNet::addSaddle(const State* const si, const State* const sj, const State* const s) {"<< std::endl;
	assertbiu(si != NULL, "si is NULL");
	assertbiu(sj != NULL, "sj is NULL");
	assertbiu(s != NULL, "s is NULL");
		
	bool changed = false;
	
	size_t 	i = getMinIndex(si);
	size_t 	j = getMinIndex(sj);
	assertbiu(i != INT_MAX, "minimum i not in topology!");
	assertbiu(j != INT_MAX, "minimum j not in topology!");
		
	//! if States are known minima.. 
	if(i != INT_MAX && j != INT_MAX) {
		//! if saddle doesn't exist OR energy of best known saddle is higher than energy of new saddle
		if((mAdjacency[i][j] == NULL) ||
					(mAdjacency[i][j]->getEnergy() > s->getEnergy())) {
			delete mAdjacency[i][j];
			mAdjacency[i][j] 	= mAdjacency[j][i] = s->clone();
			mLabel[i][j] 		= mLabel[j][i] 	= s->getEnergy();
			changed = true;
		}
	} 
	return changed;
}

const size_t 
SaddleNet::addMin(const State* const s) {
	
	assertbiu(s != NULL, "s is NULL");
	if(s==NULL) return INT_MAX;
	
	size_t 	dim = getMinCount();
	
	mLabel.resize(dim+1,dim+1,0);
	mAdjacency.resize(dim+1,dim+1,NULL);
	
	mLabel[dim][dim] = s->getEnergy();
	mAdjacency[dim][dim] = s->clone();

	return dim;
}


const State* const
SaddleNet::getMin(size_t i) const {
	assertbiu(i < mLabel.numRows(), "index i out of bound!");
	
	if(i >= mLabel.numRows()) 	return NULL;
	else						return mAdjacency[i][i];
}


const size_t 
SaddleNet::getMinIndex(const State* const s) const{
	assertbiu(s != NULL, "s is NULL");

	for (size_t i = 0; i < mLabel.numRows(); ++i)
		if(*mAdjacency[i][i] == *s) return i;
	
	return INT_MAX;
}


const size_t 
SaddleNet::getMinCount() const {
	return mLabel.numRows();
}


const std::vector<const State*> 
SaddleNet::getAllMin() const{
	std::vector<const State*> temp;
	
	for (size_t i = 0; i < mLabel.numRows(); ++i)
		temp.push_back( getMin(i) );

	return temp;
}

//! calculates the RMSD between this and a given SaddleNet(can have different # of minima)
double 
SaddleNet::getRMSD(const LandscapeTopology* const pLT) const {
	assertbiu(pLT!=NULL, "given LandscapeTopology is NULL!");
	
	const SaddleNet* const sn = dynamic_cast<const SaddleNet* const>(pLT);
	assertbiu(sn!=NULL, "given LandscapeTopology no SaddleNet!");
	
	//! net A is this, net B is given net by parameter
	double beta = (double)1. / ((double)(273.15 + 37) * (double)1.98717 );
	double rmsd = 0;
	double aMax = getMaxSaddle();
	double bMax = sn->getMaxSaddle();
	double minE; 
	
	vector<size_t> AinB;
	vector<size_t> BinA;
	
	size_t aCount = getMinCount();
	size_t bCount = sn->getMinCount();
	size_t allSaddleCount = 0;
			
	for (size_t i = 0; i < aCount; ++i)
		AinB.push_back(sn->getMinIndex(getMin(i)));
	
	for (size_t i = 0; i < bCount; ++i)
		BinA.push_back(getMinIndex(sn->getMin(i)));
	
	//! go through all saddles of saddle net A
	for (size_t i = 0; i < aCount; ++i) {
		for (size_t j = i+1; j < aCount; ++j) {
			minE = min(getMin(i)->getEnergy(), getMin(j)->getEnergy());
			//! if a minimum of net A is not found in saddle net B.. else..
			if(AinB[i]!=INT_MAX && AinB[j]!=INT_MAX) {
				++allSaddleCount;
				if(isSaddle(i,j))
					if(sn->isSaddle(AinB[i],AinB[j]))
						rmsd =+ (getSaddleVal(i,j) - getSaddleVal(AinB[i],AinB[j])) *
								(getSaddleVal(i,j) - getSaddleVal(AinB[i],AinB[j])) *
								exp(-1. * beta * minE);
					else
						rmsd =+ (getSaddleVal(i,j) - bMax) *
								(getSaddleVal(i,j) - bMax) *
								exp(-1. * beta * minE);
				else 
					//TODO complete == true angenommen
					if(/*complete &&*/ sn->isSaddle(AinB[i],AinB[j])) 
						rmsd =+ (aMax - getSaddleVal(AinB[i],AinB[j])) *
								(aMax - getSaddleVal(AinB[i],AinB[j])) *
								exp(-1. * beta * minE);
					else 
						--allSaddleCount; 
			}
			else {
				if(isSaddle(i,j)) {
					++allSaddleCount;
					//! use maximum saddleheight in B as substitution for missing minimum in B 
					rmsd =+ (getSaddleVal(i,j) - bMax) *
							(getSaddleVal(i,j) - bMax) *
							exp(-1. * beta * minE);
				}
			}
		}
	}
	//! go through the saddles of saddle net B which connect minima not known in saddle net A 
	for (size_t i = 0; i < bCount; ++i) {
		for (size_t j = i+1; j < bCount; ++j) {
			minE = min(sn->getMin(i)->getEnergy(), sn->getMin(j)->getEnergy());
			
			if(BinA[i]==INT_MAX || BinA[j]==INT_MAX) {
				++allSaddleCount;
				rmsd =+ (sn->getSaddleVal(i,j) - aMax) *
						(sn->getSaddleVal(i,j) - aMax) *
						exp(-1. * beta * minE);
			}
		}
	}
	
	//! after all divide rmsd by number of saddles
	// TODO: checken ob Andreas durch N*(N-1) oder 2N*(N-1) geteilt hat!
	rmsd = rmsd / (double)allSaddleCount;
	//! and square root..
	rmsd = sqrt(rmsd);
	
	return rmsd;
}


////////////////////////////////////////////////////////////////////////////
//////////////// SaddleNet specific members functions //////////////////////
////////////////////////////////////////////////////////////////////////////
inline
bool 
SaddleNet::isSaddle(const size_t i, const size_t j) const  {
	if(i>=mLabel.numRows() || j>=mLabel.numRows()) return false;
	return (mAdjacency[i][j]!=NULL);
}


const State* const
SaddleNet::getSaddle(const size_t i, const size_t j) const{
	assertbiu(i < mLabel.numRows(), "index i out of bound!");
	assertbiu(j < mLabel.numRows(), "index j out of bound!");
	
	if(i < mLabel.numRows() && j < mLabel.numRows()) {
		return mAdjacency[i][j];
	}
	return NULL;	
}

double
SaddleNet::getSaddleVal(const size_t i, const size_t j) const{
	assertbiu(i < mLabel.numRows(), "index i out of bound!");
	assertbiu(j < mLabel.numRows(), "index j out of bound!");
	assertbiu(i != j , "i == j for saddle between i and j!");

	if((i < mLabel.numRows() && j < mLabel.numRows()) && mAdjacency[i][j]!=NULL)
		return mLabel[i][j];
	
	return static_cast<double>(INT_MAX);
}

const double
SaddleNet::getSaddleValSum() const{
	double sum = 0;
	
	for (size_t i = 0; i < mLabel.numRows(); ++i) {
		for (size_t j = i+1; j < mLabel.numRows(); ++j) {
			if(mAdjacency[i][j]!=NULL) {
				sum += mLabel[i][j];
			}
		}
	}
	return sum;
}

const double
SaddleNet::getMediumSaddleVal() const{
	double sum = 0;
	size_t count=0;
	
	for (size_t i = 0; i < mLabel.numRows(); ++i) {
		for (size_t j = i+1; j < mLabel.numRows(); ++j) {
			if(mAdjacency[i][j]!=NULL) {
				++count;
				sum += mLabel[i][j];
			}
		}
	}
	
	if(count==0)
		return 0;
	else
		return sum/(double)count;
}

const size_t 
SaddleNet::getSaddleCount() const {
	size_t count=0;
		
	for (size_t i = 0; i < mLabel.numRows(); ++i) {
		for (size_t j = i+1; j < mLabel.numRows(); ++j) {
			if(mAdjacency[i][j]!=NULL) ++count;
		}
	}
	
	return count;
}


void 
SaddleNet::printSaddleNet() const {
//	std::cerr<<"SaddleNet "<<this<<" :"<<std::endl;
	for (size_t i = 0; i < mLabel.numRows(); ++i) {
		for (size_t j = 0; j < mLabel.numRows(); ++j) {
			std::cerr<<mLabel[i][j]<<" ";
		}
		std::cerr<<std::endl;
	}
}



double 
SaddleNet::getMaxSaddle() const {
	double max = -INT_MAX;
	
	for (size_t i = 0; i < getMinCount(); ++i) {
		for (size_t j = i+1; j < getMinCount(); ++j) {
			if(mLabel[i][j]>max) max=mLabel[i][j];
		}
	}

	return max;
}

}//ell


std::istream&
operator>>(std::istream& in,  ell::SaddleNet& sn)
{
	enum STATUS {NOTHING, MINIMUM, SADDLE};
	STATUS status;
	
	string sequence, structure, line, substring;
	stringstream sstr;
	
	size_t type= INT_MAX, dim = INT_MAX;
	size_t index_i, index_j;
	size_t lineNumber = 0;
	
	string::size_type i, k, l;
	
	State* s;
	//! first step: get the type information for the saddle net( invisible for graphviz as comments //) 
	getline(in, line);
	
	i = line.find("ELL-LT-TYPE=");
	if(i != string::npos) { 
		line=line.substr(i+12);
		
		if(line!="SN" && line!="MS") {
    		cerr << "given LT type unsupported(supported: SN, MS)!"<<endl;
    		exit(-1);
    	}
	} else  {
		cerr << "ELL-LT-TYPE missing in first line of file"<<endl;
		exit(-1);
	}
	
	getline(in, line);
	
	i = line.find("STATETYPE=");
	if(i != string::npos) { 
		line=line.substr(i+10);
		
		if(line == "RNA") type = 0;
		else {
			// TODO: keep up to date and syncronized with operator <<
			cerr << "STATETYPE "<<line<<" not supported!"<<endl;
			cerr << "possible STATETYPEs are:"<<endl;
			cerr << "RNA : S_RNAfe_SingleM"<<endl<<endl;
			cerr << "abortion in line "<<lineNumber<<" of file"<<endl;
			exit(-1);
    	} 
	} else  {
		cerr << "STATETYPE missing in second line of file"<<endl;
		exit(-1);
	}

	getline(in, line);
	i = line.find("DIMENSION=");
	if(i != string::npos) {
		line=line.substr(i+10);  		
		
		sstr.clear();
		sstr.str(line);
		sstr >> dim;
	} else  {
		cerr << "DIMENSION missing in third line of file"<<endl;
		exit(-1);
	}
	
	
	lineNumber = 3;
	//! second step: get all the minima and saddles (graphviz comments with // ) 
    while( getline(in, line) ) {
    	++lineNumber;
    	s = NULL;
    	
//    	cout << "line "<<lineNumber<<" is "<<line <<endl;
    	index_i = index_j = INT_MAX;
    	i = k = l = INT_MAX;
    	
    	status = NOTHING;
    	
    	//! if graphViz data starts.. end the reading
    	i = line.find("GRAPHVIZ");
    	if(i != string::npos) break;
     	
    	i = line.find("MINIMUM");
    	if(i != string::npos) {
    		if(type== INT_MAX || dim==INT_MAX) {
    			cerr << "wrong file format: Statetype or Dimension not defined before first state definition"<<endl;
    			cerr << "abortion in line "<<lineNumber<<endl;
    			exit(-1);
    		}
  		
    		substring = line.substr(
    				line.find("i=")+2,
    				min(line.find('('), line.find('.')) - (line.find("i=")+2));
    		sstr.clear();
    		sstr.str(substring);
    		sstr >> index_i;
//    		cout << "i= \'"<<index_i<<"\'"<<endl;
    		
    		status = MINIMUM;
    	} 
    	
    	i = line.find("SADDLE");
    	if(i != string::npos) {
    		if(type== INT_MAX || dim==INT_MAX) {
    			cerr << "wrong file format: Statetype or Dimension not defined before first state definition"<<endl;
    			cerr << "abortion in line "<<lineNumber<<endl;
    			exit(-1);
    		}

    		substring = line.substr( 
    				line.find("i=")+2 , 
    				line.find("j=") - (line.find("i=")+2));
    		sstr.clear(); 
    		sstr.str(substring);
    		sstr >> index_i;
//    		cout << "i= \'"<<index_i<<"\' ";
    		
    		substring = line.substr(
    				line.find("j=")+2 ,
    				min(line.find('('), line.find('.')) - (line.find("j=")+2));
    		sstr.clear();
    		sstr.str(substring);
    		sstr >> index_j;
//    		cout << "j= \'"<<index_j<<"\'"<<endl;
    		
    		status = SADDLE;
    	}
    	
    	if(status==NOTHING) continue;
    	
    	s = fromString(line);
    	if(s == NULL) {
    		cerr << "wrong sequence or structure string given in line "<<lineNumber<<" of file!"<<endl;
    		continue;
    	}
    	
    	//! it should be safe to assume sequence and structure are initialised
		switch(type) {
			case 0:
				if(status==MINIMUM) sn.addMin(s->clone());
				if(status==SADDLE) sn.addSaddle(index_i, index_j, s->clone());
				break;
			default: ;
		}
		
		delete s;
    }
	 
	return in;
}

std::ostream&
operator << (std::ostream& out, const ell::SaddleNet& sn) {
	size_t dim = sn.getMinCount();
	string type;

	//! first check for type of state..
	// TODO: keep up to date and syncronized with operator >>
	const S_RNAfe_SingleM* rna;
	rna = dynamic_cast<const S_RNAfe_SingleM*>(sn.getMin(0));
	if(rna!=NULL) type="RNA";
	else {
			cerr << "Statetype in BarrierTree is no RNA!"; 
			exit(-1); 
	}

	//! step 1: save the datastructure for later reconstruction as 
	//! comments in graphviz file

	out << "//ELL-LT-TYPE=SN"<< endl;
	out << "//STATETYPE="<<type<< endl;
	out << "//DIMENSION="<<dim<< endl;
	
	//! put all minima in with..
	//! //MINIMUM= index toString()-format
	for (size_t i = 0; i < dim; ++i) {
		out << "//MINIMUM\ti="<<i<<"\t"<<sn.getMin(i)->toString()<<endl;
	}

	//! put all saddles in with..
	//! //SADDLE= index_i index_j toString()-format
	for (size_t i = 0; i < dim; ++i) {
		for (size_t j = i+1; j < dim; ++j) {
			if(sn.getSaddle(i,j)!=NULL)
				out << "//SADDLE\ti="<<i<<"\tj="<<j<<"\t"<<sn.getSaddle(i,j)->toString()<<endl;
		}
	}
	
	out << endl << "//GRAPHVIZ"<<endl;
	
	//! step 2: write the graph in DOT language
	out << "graph SADDLENET {"<<endl;
	for (size_t i = 0; i < dim; ++i) {
		for (size_t j = i+1; j < dim; ++j) {
			if(sn.isSaddle(i,j))
				out << "\t\tM"<<i<<" -- M"<<j<<" [label=\""<<sn.getSaddleVal(i,j)<<"\"];"<<endl;			
		}
	}	
	out << "label = \"\\n\\nSaddleNet Diagram\\nwith "<<dim<<" minima and "
	<<sn.getSaddleCount()<<" saddles\";"<<endl;
	out << "fontsize=20;"<<endl;
	out << "} "<<endl;
	
	
	return out;
}

//std::ostream&
//operator << (std::ostream& out, const ell::SaddleNet& sn) {
//	size_t dim = sn.getMinCount();
//	
//	out << "minima:\n";
//	for (size_t i = 0; i < dim; ++i) {
//		out 
//		<<"i="<<i
//		<<"\t"<<sn.getMin(i)->getEnergy()
//		<<"\t"<<sn.getMin(i)->toString()
//		<<"\n";
//	}
//
//	out << "saddles:\n";
//	for (size_t i = 0; i < dim; ++i) {
//		for (size_t j = i+1; j < dim; ++j) {
//			if(sn.getSaddle(i,j)!=NULL) {
//				out 
//				<<"i="<<i
//				<<"\tj="<<j
//				<<"\t"<<sn.getSaddleVal(i,j)
//				<<"\t"<<sn.getSaddle(i,j)->toString()
//				<<"\n";
//			}
//		}
//	}
//	out.flush();
//
//	return out;
//}



