#ifndef SADDLENET_HH_
#define SADDLENET_HH_

#include "ell/tp/LandscapeTopology.hh"
#include <ell/State.hh>
#include <biu/Matrix.hh>
#include <vector>

namespace ell {
	 

class SaddleNet: public LandscapeTopology {
protected: 
	biu::Matrix<State*>		 mAdjacency;
	biu::Matrix<double>		 mLabel;

public:
	static	size_t		rejSaddles;
	static	size_t		addSaddles;
	
	virtual ~SaddleNet();
	
	///////////////// LANDSCAPE TOPOLOGY INTERFACE //////////////////
	//! sets a Saddle State between Minima with indices i and j
	virtual bool 
	addSaddle(const size_t i, const size_t j, const State* const s);
	
	//! sets a Saddle State between Minima States i and j
	virtual bool 
	addSaddle(const State* const i, const State* const j, const State* const s);
	
	//! adds a minimum State and returns the index of this minimum 
	virtual const size_t 
	addMin(const State* const i);
	
	//! returns index of State if it's a minimum, returns INT_MAX if State is no known minimum 
	virtual const size_t 
	getMinIndex(const State* const i) const;
	
	//! returns number of stored minima
	virtual const size_t 
	getMinCount() const;
	
	//! returns State of minimum with given index, returns NULL if index is out of bounds
	virtual const State* const 
	getMin(size_t i) const;
	
	//! returns the internal vector of minima as a const reference
	virtual const std::vector<const State*> 
	getAllMin() const;
	
	//! calculates the RMSD between this and a given SaddleNet(can have different # of minima)
	virtual double 
	getRMSD(const LandscapeTopology* const) const;
	
	///////////////// ADDITIONAL SADDLE NET INTERFACE //////////////////	
	inline
	bool 
	isSaddle(const size_t i, const size_t j) const;
	
	//! returns the State of the Saddle between minima mit indices i and j and NULL if there is no saddle
	const State* const 
	getSaddle(const size_t i, const size_t j) const;
		
	//! returns energy of Saddle between minima mit indices i and j
	double 
	getSaddleVal(const size_t, const size_t) const;

	//! returns medium energy of all saddles
	const double
	getMediumSaddleVal() const;
	
	//! returns the energy sum of all saddles
	const double
	getSaddleValSum() const;
	
	//! returns the number of all saddles
	const size_t
	getSaddleCount() const;

	//! prints the SaddleNet
	void 
	printSaddleNet() const;

	//! gets maximal saddle height
	double 
	getMaxSaddle() const;
	
};

}//ell


//! Reads the elements of the network(minima, saddle heights) from a stream.
//! @param in the input stream to read from
//! @param sn the SaddleNet to fill
//! @return the read input stream 
std::istream&
operator>>(std::istream& in,  ell::SaddleNet& sn);

//! Prints the saddlenet with structure und saddleenergies to stream
//! in graphviz format, see http://www.graphviz.org
//! @param out the output stream to write to
//! @param sn the SaddleNet to print
//! @return the changed output stream 'out'
std::ostream&
operator << (std::ostream& out, const ell::SaddleNet& sn);

////! Prints the saddlenet as pairs of Saddleindices and energy to outstream
////! @param out the output stream to write to
////! @param sn the SaddleNet to print
////! @return the changed output stream 'out'
//std::ostream&
//operator << (std::ostream& out, const ell::SaddleNet& sn);


#endif /*SADDLENET_HH_*/
